# Original work Copyright 2017 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import os
import sys
import threading

from contextlib import contextmanager
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.options import DEFAULT_TIMEOUT
from typing import TypeVar, Any, Callable, Tuple
from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled
from functools import lru_cache


PARENT_PROCESS_WATCH_INTERVAL = 3  # 3 s


def as_str(s) -> str:
    if isinstance(s, bytes):
        return s.decode("utf-8", "replace")
    return str(s)


log = get_logger(__name__)


def list_to_string(value):
    return ",".join(value) if isinstance(value, list) else value


if sys.platform == "win32":
    import ctypes

    kernel32 = ctypes.windll.kernel32
    PROCESS_SYNCHRONIZE = 0x00100000
    DWORD = ctypes.c_uint32
    BOOL = ctypes.c_int
    LPVOID = ctypes.c_void_p
    HANDLE = LPVOID

    OpenProcess = kernel32.OpenProcess
    OpenProcess.argtypes = [DWORD, BOOL, DWORD]
    OpenProcess.restype = HANDLE

    WaitForSingleObject = kernel32.WaitForSingleObject
    WaitForSingleObject.argtypes = [HANDLE, DWORD]
    WaitForSingleObject.restype = DWORD

    WAIT_TIMEOUT = 0x00000102
    WAIT_ABANDONED = 0x00000080
    WAIT_OBJECT_0 = 0
    WAIT_FAILED = 0xFFFFFFFF

    def is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Running `os.kill()` on Windows always exits the process, so it can't be used to check for an alive process.
        see: https://docs.python.org/3/library/os.html?highlight=os%20kill#os.kill

        Hence ctypes is used to check for the process directly via windows API avoiding any other 3rd-party dependency.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        process = OpenProcess(PROCESS_SYNCHRONIZE, 0, pid)
        if process != 0:
            try:
                wait_result = WaitForSingleObject(process, 0)
                if wait_result == WAIT_TIMEOUT:
                    return True
            finally:
                kernel32.CloseHandle(process)
        return False


else:
    import errno

    def _is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return False  # No such process.
            elif e.errno == errno.EPERM:
                return True  # permission denied.
            else:
                log.info("Unexpected errno: %s", e.errno)
                return False
        else:
            return True

    def is_process_alive(pid):
        from robocorp_ls_core.subprocess_wrapper import subprocess

        if _is_process_alive(pid):
            # Check if zombie...
            try:
                cmd = ["ps", "-p", str(pid), "-o", "stat"]
                try:
                    process = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                except:
                    log.exception("Error calling: %s.", " ".join(cmd))
                else:
                    stdout, _ = process.communicate()
                    stdout = stdout.decode("utf-8", "replace")
                    lines = [line.strip() for line in stdout.splitlines()]
                    if len(lines) > 1:
                        if lines[1].startswith("Z"):
                            return False  # It's a zombie
            except:
                log.exception("Error checking if process is alive.")

            return True
        return False


def _popen(cmdline, **kwargs):
    import subprocess

    try:
        return subprocess.Popen(cmdline, **kwargs)
    except:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _call(cmdline, **kwargs):
    import subprocess

    try:
        subprocess.check_call(cmdline, **kwargs)
    except:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _kill_process_and_subprocess_linux(pid):
    import subprocess

    initial_pid = pid

    def list_children_and_stop_forking(ppid):
        children_pids = []
        _call(
            ["kill", "-STOP", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        list_popen = _popen(
            ["pgrep", "-P", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        if list_popen is not None:
            stdout, _ = list_popen.communicate()
            for line in stdout.splitlines():
                line = line.decode("ascii").strip()
                if line:
                    pid = str(line)
                    children_pids.append(pid)
                    # Recursively get children.
                    children_pids.extend(list_children_and_stop_forking(pid))
        return children_pids

    previously_found = set()

    for _ in range(50):  # Try this at most 50 times before giving up.
        children_pids = list_children_and_stop_forking(initial_pid)
        found_new = False

        for pid in children_pids:
            if pid not in previously_found:
                found_new = True
                previously_found.add(pid)
                _call(
                    ["kill", "-KILL", str(pid)],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

        if not found_new:
            break

    # Now, finish the initial one.
    _call(
        ["kill", "-KILL", str(initial_pid)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


def kill_process_and_subprocesses(pid):
    log.debug("Killing process and subprocesses of: %s", pid)
    from subprocess import CalledProcessError

    if sys.platform == "win32":
        from robocorp_ls_core.subprocess_wrapper import subprocess

        args = ["taskkill", "/F", "/PID", str(pid), "/T"]
        retcode = subprocess.call(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE
        )
        if retcode not in (0, 128, 255):
            raise CalledProcessError(retcode, args)
    else:
        _kill_process_and_subprocess_linux(pid)


_track_pids_to_exit = set()
_watching_thread_global = None


def exit_when_pid_exists(pid):
    _track_pids_to_exit.add(pid)
    global _watching_thread_global
    if _watching_thread_global is None:
        import time

        def watch_parent_process():
            # exit when any of the ids we're tracking exit.
            while True:
                for pid in _track_pids_to_exit:
                    if not is_process_alive(pid):
                        # Note: just exit since the parent process already
                        # exited.
                        log.info(
                            f"Force-quit process: %s because parent: %s exited",
                            os.getpid(),
                            pid,
                        )
                        os._exit(0)

                time.sleep(PARENT_PROCESS_WATCH_INTERVAL)

        _watching_thread_global = threading.Thread(target=watch_parent_process, args=())
        _watching_thread_global.daemon = True
        _watching_thread_global.start()


F = TypeVar("F", bound=Callable[..., Any])


def overrides(method: Any) -> Callable[[F], F]:
    """
    Meant to be used as

    class B:
        @overrides(A.m1)
        def m1(self):
            pass
    """

    @functools.wraps(method)
    def wrapper(func):
        if func.__name__ != method.__name__:
            msg = f"Wrong @override: {func.__name__!r} expected, but overwriting {method.__name__!r}."
            raise AssertionError(msg)

        return func

    return wrapper


def implements(method: Any) -> Callable[[F], F]:
    @functools.wraps(method)
    def wrapper(func):
        if func.__name__ != method.__name__:
            msg = f"Wrong @implements: {func.__name__!r} expected, but implementing {method.__name__!r}."
            raise AssertionError(msg)

        return func

    return wrapper


def log_and_silence_errors(logger, return_on_error=None):
    def inner(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except JsonRpcRequestCancelled:
                logger.info("Cancelled handling: %s", func)
                raise  # Don't silence cancelled exceptions
            except:
                logger.exception("Error calling: %s", func)
                return return_on_error

        return new_func

    return inner


@contextmanager
def after(obj, method_name, callback):
    original_method = getattr(obj, method_name)

    @functools.wraps(original_method)
    def new_method(*args, **kwargs):
        ret = original_method(*args, **kwargs)
        callback(*args, **kwargs)
        return ret

    setattr(obj, method_name, new_method)
    try:
        yield
    finally:
        setattr(obj, method_name, original_method)


@contextmanager
def before(obj, method_name, callback):
    original_method = getattr(obj, method_name)

    @functools.wraps(original_method)
    def new_method(*args, **kwargs):
        callback(*args, **kwargs)
        ret = original_method(*args, **kwargs)
        return ret

    setattr(obj, method_name, new_method)
    try:
        yield
    finally:
        setattr(obj, method_name, original_method)


def check_min_version(version: str, min_version: Tuple[int, int]) -> bool:
    """
    :param version:
        This is the version of robotframework.

    :param min_version:
        This is the minimum version to match.

    :return bool:
        True if version >= min_versiond and False otherwise.
    """
    try:
        v = tuple(int(x) for x in version.split("."))
    except:
        return False

    return v >= min_version


def wait_for_condition(condition, msg=None, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0):
    import time

    curtime = time.time()

    while True:
        if condition():
            break
        if timeout is not None and (time.time() - curtime > timeout):
            error_msg = f"Condition not reached in {timeout} seconds"
            if msg is not None:
                error_msg += "\n"
                if callable(msg):
                    error_msg += msg()
                else:
                    error_msg += str(msg)

            raise TimeoutError(error_msg)
        time.sleep(sleep)


def wait_for_non_error_condition(
    generate_error_or_none, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0
):
    import time

    curtime = time.time()

    while True:
        error_msg = generate_error_or_none()
        if error_msg is None:
            break

        if timeout is not None and (time.time() - curtime > timeout):
            raise TimeoutError(
                f"Condition not reached in {timeout} seconds\n{error_msg}"
            )
        time.sleep(sleep)


def wait_for_expected_func_return(
    func, expected_return, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0
):
    def check():
        found = func()
        if found != expected_return:
            return "Expected: %s. Found: %s" % (expected_return, found)

        return None

    wait_for_non_error_condition(check, timeout, sleep)


def isinstance_name(obj, classname, memo={}):
    """
    Checks if a given object is instance of a class with the given name.
    """
    if classname.__class__ in (list, tuple):
        for c in classname:
            if isinstance_name(obj, c):
                return True
        return False

    cls = obj.__class__
    key = (cls, classname)
    try:
        return memo[key]
    except KeyError:
        if cls.__name__ == classname:
            memo[key] = True
        else:
            for check in obj.__class__.__mro__:
                if check.__name__ == classname:
                    memo[key] = True
                    break
            else:
                memo[key] = False

        return memo[key]


def build_subprocess_kwargs(cwd, env, **kwargs) -> dict:
    from robocorp_ls_core.subprocess_wrapper import subprocess

    startupinfo = None
    if sys.platform == "win32":
        # We don't want to show the shell on windows!
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE
        startupinfo = startupinfo

    if cwd:
        kwargs["cwd"] = cwd
    if env:
        kwargs["env"] = env
    kwargs["startupinfo"] = startupinfo
    return kwargs


def make_unique(lst):
    seen = set()
    return [x for x in lst if x not in seen and not seen.add(x)]


@lru_cache(maxsize=3000)
def normalize_filename(filename):
    return os.path.abspath(os.path.normpath(os.path.normcase(filename)))



from robocorp_ls_core.robotframework_log import get_logger
import os
from robocorp_ls_core.protocols import IEndPoint, IDirCache
from typing import (
    Optional,
    List,
    Any,
    Tuple,
    Callable,
    Iterable,
    Dict,
    Iterator,
    Set,
)
from robocorp_ls_core.constants import NULL
from functools import partial
import threading
from robotframework_ls.impl.robot_lsp_constants import CHECK_IF_LIBRARIES_INSTALLED
from robocorp_ls_core.basic import normalize_filename

log = get_logger(__name__)


def _normfile(filename):
    return normalize_filename(filename)


def _norm_filename(path):
    return normalize_filename(os.path.realpath(path))


class _Node(object):
    def __init__(self, name: str) -> None:
        self.name = name
        self._children: Dict[str, _Node] = {}
        self.is_leaf_libname = False
        self.full_libname = ""

    def __str__(self):
        return f"_Node({self.name})"

    __repr__ = __str__

    def add_child(self, name: str) -> "_Node":
        c = self._children.get(name)
        if c is None:  # If it exists, just reuse it.
            c = self._children[name] = _Node(name)
        return c

    def get_child(self, name: str) -> Optional["_Node"]:
        return self._children.get(name)

    def has_children(self):
        return len(self._children) > 0


class LibspecWarmup(object):
    """
    This class pre-generates the .libspec files as needed.

    Note that it's a "friend" of LibspecManager and may access its internal apis.
    """

    def __init__(
        self,
        endpoint: Optional[IEndPoint] = None,
        dir_cache: Optional[IDirCache] = None,
    ):
        self._endpoint = endpoint
        self._dir_cache = dir_cache

    def _generate(
        self,
        libspec_manager,
        progress_title: str,
        elapsed_time_key: Optional[str],
        mutex_name_prefix: str,
        on_finish: Callable[[], Any],
        provide_libname_and_kwargs_to_create_libspec: Callable[
            [], Iterable[Tuple[str, dict]]
        ],
    ):
        from robocorp_ls_core.progress_report import progress_context
        from robocorp_ls_core.progress_report import ProgressWrapperForTotalWork

        try:
            ctx: Any
            if self._endpoint is not None:
                ctx = progress_context(
                    self._endpoint,
                    progress_title,
                    self._dir_cache,
                    elapsed_time_key=elapsed_time_key,
                )
            else:
                ctx = NULL

            import time
            from concurrent import futures
            from robocorp_ls_core.system_mutex import timed_acquire_mutex
            from robocorp_ls_core.system_mutex import generate_mutex_name

            initial_time = time.time()
            wait_for = []

            max_workers = min(10, (os.cpu_count() or 1) + 4)
            thread_pool = futures.ThreadPoolExecutor(max_workers=max_workers)

            try:
                log.debug(f"Waiting for mutex to {progress_title}.")
                with timed_acquire_mutex(
                    generate_mutex_name(
                        _norm_filename(libspec_manager._builtins_libspec_dir),
                        prefix=mutex_name_prefix,
                    ),
                    timeout=100,
                ):
                    library_names = set(libspec_manager.get_library_names())

                    with ctx as progress_reporter:
                        progress_wrapper = ProgressWrapperForTotalWork(
                            progress_reporter
                        )

                        log.debug(f"Obtained mutex to {progress_title}.")
                        for (
                            libname,
                            kwargs,
                        ) in provide_libname_and_kwargs_to_create_libspec():
                            libspec_filename = (
                                libspec_manager._compute_libspec_filename(
                                    libname, **kwargs
                                )
                            )
                            if (
                                os.path.exists(libspec_filename)
                                or libname in library_names
                            ):
                                continue
                            progress_wrapper.increment_total_steps()

                            def progress_and_create(libname, kwargs):
                                libspec_manager._create_libspec(libname, **kwargs)
                                progress_wrapper.increment_step_done()

                            wait_for.append(
                                thread_pool.submit(
                                    partial(progress_and_create, libname, kwargs)
                                )
                            )

                        for future in wait_for:
                            future.result()

                        if wait_for:
                            log.debug(
                                "Total time to %s: %.2fs"
                                % (progress_title, time.time() - initial_time)
                            )
                            on_finish()
            finally:
                thread_pool.shutdown(wait=False)
        except:
            log.exception(f"Error {progress_title}.")
        finally:
            log.debug(f"Finished {progress_title}.")

    def gen_builtin_libraries(self, libspec_manager) -> None:
        """
        Generates .libspec files for the libraries builtin (if needed).
        """
        from robotframework_ls.impl import robot_constants
        from robotframework_ls.impl.robot_constants import RESERVED_LIB

        def provide_libname_and_kwargs_to_create_libspec():
            for libname in robot_constants.STDLIBS:
                if libname == RESERVED_LIB:
                    continue
                yield libname, dict(is_builtin=True)

        self._generate(
            libspec_manager,
            progress_title="Generate .libspec for builtin libraries",
            elapsed_time_key="generate_builtins_libspec",
            mutex_name_prefix="gen_builtins_",
            on_finish=lambda: None,
            provide_libname_and_kwargs_to_create_libspec=provide_libname_and_kwargs_to_create_libspec,
        )

    def find_rf_libraries(
        self,
        libspec_manager,
        target_libs: Iterable[str] = CHECK_IF_LIBRARIES_INSTALLED,
        tracked_folders=None,
    ) -> Set[str]:
        """
        Given an input of target libraries, try to discover if those are available
        in the PYTHONPATH.
        """

        root = _Node("")
        for libname in target_libs:
            splitted = libname.split(".")
            assert splitted
            parent = root

            for name in splitted:
                parent = parent.add_child(name)

            # The last one is marked as being an actual entry.
            parent.is_leaf_libname = True
            parent.full_libname = libname

        found_libnames: Set[str] = set()
        if tracked_folders is None:
            tracked_folders = libspec_manager.collect_all_tracked_folders()
        for s in tracked_folders:
            if os.path.isdir(s):
                found_libnames.update(self._veriy_tree_match(root, s, True))

        return found_libnames

    def _veriy_tree_match(
        self, parent: _Node, path: str, is_dir: bool
    ) -> Iterator[str]:
        try:
            if parent.has_children():
                if is_dir:
                    for dir_entry in os.scandir(path):
                        if dir_entry.is_dir():
                            use_name = dir_entry.name
                        else:
                            use_name = dir_entry.name
                            if not use_name.endswith(".py"):
                                continue
                            use_name = os.path.splitext(use_name)[0]

                        child_node = parent.get_child(use_name)
                        if child_node is not None:
                            if child_node.is_leaf_libname:
                                yield child_node.full_libname

                            yield from self._veriy_tree_match(
                                child_node, dir_entry.path, dir_entry.is_dir()
                            )
        except:
            log.exception(
                f"Exception handled while computing preinstalled libraries when verifying: {path}"
            )

    def gen_user_libraries(self, libspec_manager, user_libraries: List[str]):
        """
        Pre-generate the libspec for user libraries installed (in a thread).

        Note that it'll do it for the libraries that the user pre-specifies as well
        as those we automatically find in the PYTHONPATH.

        :param user_libraries:
            The libraries that the user specifies.
        """

        def provide_libname_and_kwargs_to_create_libspec():
            for libname in user_libraries:
                yield libname, {}

            if os.environ.get(
                "ROBOTFRAMEWORK_LS_PRE_GENERATE_PYTHONPATH_LIBS", ""
            ).lower() not in (
                "0",
                "false",
            ):
                # Now, besides those (specified by the user), we'll also try to find
                # existing libraries in the PYTHONPATH. Note that it's possible to
                # set an environment variable such as:
                # ROBOTFRAMEWORK_LS_PRE_GENERATE_PYTHONPATH_LIBS=0
                # to skip this process.
                for libname in self.find_rf_libraries(libspec_manager):
                    yield libname, {}

        def in_thread():
            self._generate(
                libspec_manager,
                progress_title="Generate .libspec for libraries",
                elapsed_time_key="generate_libspec",
                mutex_name_prefix="gen_libspec_",
                on_finish=libspec_manager.synchronize_internal_libspec_folders,
                provide_libname_and_kwargs_to_create_libspec=provide_libname_and_kwargs_to_create_libspec,
            )

        t = threading.Thread(target=in_thread)
        t.daemon = True
        t.start()



"""
Documentation checkers
"""
from robot.parsing.model.blocks import SettingSection
from robot.parsing.model.statements import Documentation

from robocop.checkers import VisitorChecker
from robocop.rules import Rule, RuleSeverity

rules = {
    "0201": Rule(
        rule_id="0201",
        name="missing-doc-keyword",
        msg="Missing documentation in '{{ name }}' keyword",
        severity=RuleSeverity.WARNING,
        docs="""
        You can add documentation to keyword using following syntax::
        
            Keyword
                [Documentation]  Keyword documentation
                Keyword Step
                Other Step

        """,
    ),
    "0202": Rule(
        rule_id="0202",
        name="missing-doc-test-case",
        msg="Missing documentation in '{{ name }}' test case",
        severity=RuleSeverity.WARNING,
        docs="""
        You can add documentation to test case using following syntax::
        
            Test
                [Documentation]  Test documentation
                Keyword Step
                Other Step
        
        """,
    ),
    "0203": Rule(
        rule_id="0203",
        name="missing-doc-suite",
        msg="Missing documentation in suite",
        severity=RuleSeverity.WARNING,
        docs="""
        You can add documentation to suite using following syntax::
        
            *** Settings ***
            Documentation    Suite documentation

        """,
    ),
}


class MissingDocumentationChecker(VisitorChecker):
    """Checker for missing documentation."""

    reports = (
        "missing-doc-keyword",
        "missing-doc-test-case",
        "missing-doc-suite",
    )

    def visit_Keyword(self, node):  # noqa
        if node.name.lstrip().startswith("#"):
            return
        self.check_if_docs_are_present(node, "missing-doc-keyword")

    def visit_TestCase(self, node):  # noqa
        self.check_if_docs_are_present(node, "missing-doc-test-case")

    def visit_SettingSection(self, node):  # noqa
        self.check_if_docs_are_present(node, "missing-doc-suite")

    def visit_File(self, node):  # noqa
        for section in node.sections:
            if isinstance(section, SettingSection):
                break
        else:
            self.report("missing-doc-suite", node=node, lineno=1)
        super().visit_File(node)

    def check_if_docs_are_present(self, node, msg):
        for statement in node.body:
            if isinstance(statement, Documentation):
                break
        else:
            if hasattr(node, "name"):
                self.report(msg, name=node.name, node=node)
            else:
                self.report(msg, node=node)



class LoopHolder:

    @staticmethod
    def gui_loop():
        print('gui_loop() called')


def call_method():
    from _pydevd_bundle.pydevd_constants import get_global_debugger
    py_db = get_global_debugger()

    # Check state prior to breaking
    assert not py_db.gui_in_use
    assert py_db._installed_gui_support
    assert py_db._gui_event_loop == '__main__.LoopHolder.gui_loop'

    print('break here')

    assert py_db.gui_in_use
    assert py_db._installed_gui_support
    assert py_db._gui_event_loop == '__main__.LoopHolder.gui_loop'


if __name__ == '__main__':
    call_method()
    print('TEST SUCEEDED!')



if __name__ == '__main__':
    import sys
    import os
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))

    from not_my_code import other_noop
    other_noop.call_noop()  # break here
    print('TEST SUCEEDED!')



def bar():
    print('on bar mark')


def call_outer(*args):
    print('on outer mark')


def foo(*args):
    print('on foo mark')


def main():
    call_outer(foo(bar()))  # break here


if __name__ == '__main__':
    main()

    print('TEST SUCEEDED')



from typing import Optional

from robocorp_ls_core.ep_resolve_interpreter import IInterpreterInfo
from robocorp_ls_core.protocols import IConfig


def apply_interpreter_info_to_config(
    config: IConfig, interpreter_info: Optional[IInterpreterInfo]
):
    from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHON_ENV
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_PYTHON_EXECUTABLE,
    )
    from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHONPATH

    if interpreter_info is not None:
        overridden_settings: dict = {}
        python_exe = interpreter_info.get_python_exe()
        if python_exe:
            overridden_settings[OPTION_ROBOT_PYTHON_EXECUTABLE] = python_exe

        environ = interpreter_info.get_environ()
        if environ:
            overridden_settings[OPTION_ROBOT_PYTHON_ENV] = environ

        additional_pythonpath_entries = (
            interpreter_info.get_additional_pythonpath_entries()
        )
        if additional_pythonpath_entries:
            overridden_settings[OPTION_ROBOT_PYTHONPATH] = additional_pythonpath_entries
        else:
            # If we're applying the interpreter info, the pythonpath should be given
            # by it and if not available it should be overridden.
            overridden_settings[OPTION_ROBOT_PYTHONPATH] = []

        config.set_override_settings(overridden_settings)
    else:
        config.set_override_settings({})



import time
import multiprocessing
import threading
import sys
import os

event = threading.Event()


class MyThread(threading.Thread):

    def run(self):
        _a = 10
        _b = 20  # break in thread here
        event.set()
        _c = 20


def run_in_multiprocess():
    _a = 30
    _b = 40  # break in process here


if __name__ == '__main__':
    MyThread().start()
    event.wait()
    if sys.version_info[0] >= 3 and sys.platform != 'win32':
        multiprocessing.set_start_method('fork')
    p = multiprocessing.Process(target=run_in_multiprocess, args=())
    p.start()
    print('TEST SUCEEDED!')  # break in main here
    p.join()



import asyncio


async def count():
    print('enter count')
    await asyncio.sleep(.001)  # break count 1
    await asyncio.sleep(.001)  # break count 2


async def count2():
    print('enter count 2')
    await asyncio.sleep(.001)
    await asyncio.sleep(.001)


async def count3():
    print('enter count 3')
    await asyncio.sleep(.001)
    await asyncio.sleep(.001)


async def main():
    await count()  # break main
    await count2()  # step main
    await count3()


if __name__ == "__main__":
    if hasattr(asyncio, 'run'):
        print('using asyncio.run')
        asyncio.run(main())
    else:
        print('using loop.run_until_complete')
        loop = asyncio.get_event_loop()
        loop.run_until_complete(main())
        loop.close()
    print('TEST SUCEEDED!')



from _pydevd_bundle.pydevd_constants import IS_WINDOWS


def test_in_project_roots_prefix_01(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering
    files_filtering = FilesFiltering()

    another = str(tmpdir.join('another'))
    assert not another.endswith('/') and not another.endswith('\\')

    files_filtering.set_library_roots([another])
    files_filtering.set_project_roots([])
    assert not files_filtering.in_project_roots(another + '/f.py')
    if IS_WINDOWS:
        assert not files_filtering.in_project_roots(another + '\\f.py')
    else:
        assert files_filtering.in_project_roots(another + '\\f.py')

    assert files_filtering.in_project_roots(another + 'f.py')


def test_in_project_roots_prefix_02(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering
    files_filtering = FilesFiltering()

    another = str(tmpdir.join('another'))
    assert not another.endswith('/') and not another.endswith('\\')

    files_filtering.set_library_roots([])
    files_filtering.set_project_roots([another])
    assert files_filtering.in_project_roots(another + '/f.py')
    if IS_WINDOWS:
        assert files_filtering.in_project_roots(another + '\\f.py')
    else:
        assert not files_filtering.in_project_roots(another + '\\f.py')

    assert not files_filtering.in_project_roots(another + 'f.py')


def test_in_project_roots(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering
    files_filtering = FilesFiltering()

    import os.path
    import sys
    assert files_filtering._get_library_roots() == [
        os.path.normcase(x) + ('\\' if IS_WINDOWS else '/') for x in files_filtering._get_default_library_roots()]

    site_packages = tmpdir.mkdir('site-packages')
    project_dir = tmpdir.mkdir('project')

    project_dir_inside_site_packages = str(site_packages.mkdir('project'))
    site_packages_inside_project_dir = str(project_dir.mkdir('site-packages'))

    # Convert from pytest paths to str.
    site_packages = str(site_packages)
    project_dir = str(project_dir)
    tmpdir = str(tmpdir)

    # Test permutations of project dir inside site packages and vice-versa.
    files_filtering.set_project_roots([project_dir, project_dir_inside_site_packages])
    files_filtering.set_library_roots([site_packages, site_packages_inside_project_dir])

    check = [
        (tmpdir, False),
        (site_packages, False),
        (site_packages_inside_project_dir, False),
        (project_dir, True),
        (project_dir_inside_site_packages, True),
    ]
    for (check_path, find) in check[:]:
        filename_inside = os.path.join(check_path, 'a.py')
        with open(filename_inside, 'w') as stream:
            # Note: on Github actions, tmpdir may be something as:
            # c:\\users\\runner~1\\appdata\\local\\temp\\pytest-of-runneradmin\\pytest-0\\test_in_project_roots0
            # internally this may be set as:
            # c:\\users\\runneradmin\\appdata\\local\\temp\\pytest-of-runneradmin\\pytest-0\\test_in_project_roots0
            # So, when getting the absolute path, `runner~1` will be properly expanded to `runneradmin` if the
            # file exists, but if it doesn't it's not (which may make the test fail), so, make sure
            # that we actually create the file so that things work as expected.
            stream.write('...')
        check.append((filename_inside, find))

    for check_path, find in check:
        if files_filtering.in_project_roots(check_path) != find:
            if find:
                msg = 'Expected %s to be in the project roots.\nProject roots: %s\nLibrary roots: %s\n'
            else:
                msg = 'Expected %s NOT to be in the project roots.\nProject roots: %s\nLibrary roots: %s\n'

            raise AssertionError(msg % (
                check_path,
                files_filtering._get_project_roots(),
                files_filtering._get_library_roots(),
                )
            )

    files_filtering.set_project_roots([])
    files_filtering.set_library_roots([site_packages, site_packages_inside_project_dir])

    # If the IDE did not set the project roots, consider anything not in the site
    # packages as being in a project root (i.e.: we can calculate default values for
    # site-packages but not for project roots).
    check = [
        (tmpdir, True),
        (site_packages, False),
        (site_packages_inside_project_dir, False),
        (project_dir, True),
        (project_dir_inside_site_packages, False),
        ('<foo>', False),
        ('<ipython>', True),
        ('<frozen importlib._bootstrap>', False),
    ]

    for check_path, find in check:
        assert files_filtering.in_project_roots(check_path) == find, \
            'Expected: %s to be a part of the project: %s' % (check_path, find)

    sys.path.append(str(site_packages))
    try:
        default_library_roots = files_filtering._get_default_library_roots()
        assert len(set(default_library_roots)) == len(default_library_roots), \
            'Duplicated library roots found in: %s' % (default_library_roots,)

        assert str(site_packages) in default_library_roots
        for path in sys.path:
            if os.path.exists(path) and path.endswith('site-packages'):
                assert path in default_library_roots
    finally:
        sys.path.remove(str(site_packages))


def test_filtering(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering
    from _pydevd_bundle.pydevd_filtering import ExcludeFilter
    files_filtering = FilesFiltering()

    site_packages = tmpdir.mkdir('site-packages')
    project_dir = tmpdir.mkdir('project')

    project_dir_inside_site_packages = str(site_packages.mkdir('project'))
    site_packages_inside_project_dir = str(project_dir.mkdir('site-packages'))

    files_filtering.set_exclude_filters([
        ExcludeFilter('**/project*', True, True),
        ExcludeFilter('**/bar*', False, True),
    ])
    assert files_filtering.exclude_by_filter('/foo/project', None) is True
    assert files_filtering.exclude_by_filter('/foo/unmatched', None) is None
    assert files_filtering.exclude_by_filter('/foo/bar', None) is False


def test_glob_matching():
    from _pydevd_bundle.pydevd_filtering import glob_matches_path

    # Linux
    for sep, altsep in (('\\', '/'), ('/', None)):

        def build(path):
            if sep == '/':
                return path
            else:
                return ('c:' + path).replace('/', '\\')

        assert glob_matches_path(build('/a'), r'*', sep, altsep)

        assert not glob_matches_path(build('/a/b/c/some.py'), '/a/**/c/so?.py', sep, altsep)

        assert glob_matches_path('/a/b/c', '/a/b/*')
        assert not glob_matches_path('/a/b', '/*')
        assert glob_matches_path('/a/b', '/*/b')
        assert glob_matches_path('/a/b', '**/*')
        assert not glob_matches_path('/a/b', '**/a')

        assert glob_matches_path(build('/a/b/c/d'), '**/d', sep, altsep)
        assert not glob_matches_path(build('/a/b/c/d'), '**/c', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '**/c/d', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '**/b/c/d', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '/*/b/*/d', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '**/c/*', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '/a/**/c/*', sep, altsep)

        # I.e. directories are expected to end with '/', so, it'll match
        # something as **/directory/**
        assert glob_matches_path(build('/a/b/c/'), '**/c/**', sep, altsep)
        assert glob_matches_path(build('/a/b/c/'), '**/c/', sep, altsep)
        # But not something as **/directory (that'd be a file match).
        assert not glob_matches_path(build('/a/b/c/'), '**/c', sep, altsep)
        assert not glob_matches_path(build('/a/b/c'), '**/c/', sep, altsep)

        assert glob_matches_path(build('/a/b/c/d.py'), '/a/**/c/*', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d.py'), '/a/**/c/*.py', sep, altsep)
        assert glob_matches_path(build('/a/b/c/some.py'), '/a/**/c/so*.py', sep, altsep)
        assert glob_matches_path(build('/a/b/c/some.py'), '/a/**/c/som?.py', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '/**', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d'), '/**/d', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d.py'), '/**/*.py', sep, altsep)
        assert glob_matches_path(build('/a/b/c/d.py'), '**/c/*.py', sep, altsep)

        if IS_WINDOWS:
            assert glob_matches_path(build('/a/b/c/d.py'), '**/C/*.py', sep, altsep)
            assert glob_matches_path(build('/a/b/C/d.py'), '**/c/*.py', sep, altsep)

        # Expected not to match.
        assert not glob_matches_path(build('/a/b/c/d'), '/**/d.py', sep, altsep)
        assert not glob_matches_path(build('/a/b/c/d.pyx'), '/a/**/c/*.py', sep, altsep)
        assert not glob_matches_path(build('/a/b/c/d'), '/*/d', sep, altsep)

        if sep == '/':
            assert not glob_matches_path(build('/a/b/c/d'), r'**\d', sep, altsep)  # Match with \ doesn't work on linux...
            assert not glob_matches_path(build('/a/b/c/d'), r'c:\**\d', sep, altsep)  # Match with drive doesn't work on linux...
        else:
            # Works in Windows.
            assert glob_matches_path(build('/a/b/c/d'), r'**\d', sep, altsep)
            assert glob_matches_path(build('/a/b/c/d'), r'c:\**\d', sep, altsep)

        # Corner cases
        assert not glob_matches_path(build('/'), r'', sep, altsep)
        assert glob_matches_path(build(''), r'', sep, altsep)
        assert not glob_matches_path(build(''), r'**', sep, altsep)
        assert glob_matches_path(build('/'), r'**', sep, altsep)
        assert glob_matches_path(build('/'), r'*', sep, altsep)


def test_rules_to_exclude_filter(tmpdir):
    from _pydevd_bundle.pydevd_process_net_command_json import _convert_rules_to_exclude_filters
    from _pydevd_bundle.pydevd_filtering import ExcludeFilter
    from random import shuffle
    dira = tmpdir.mkdir('a')
    dirb = dira.mkdir('b')
    fileb = dirb.join('fileb.py')
    fileb2 = dirb.join('fileb2.py')
    with fileb.open('w') as stream:
        stream.write('')

    def on_error(msg):
        raise AssertionError(msg)

    rules = [
        {'path': str(dira), 'include': False},
        {'path': str(dirb), 'include': True},
        {'path': str(fileb), 'include': True},
        {'path': str(fileb2), 'include': True},
        {'path': '**/foo/*.py', 'include': True},
        {'module': 'bar', 'include': False},
        {'module': 'bar.foo', 'include': True},
    ]
    shuffle(rules)
    exclude_filters = _convert_rules_to_exclude_filters(rules, on_error)
    assert exclude_filters == [
        ExcludeFilter(name=str(fileb2), exclude=False, is_path=True),
        ExcludeFilter(name=str(fileb), exclude=False, is_path=True),
        ExcludeFilter(name=str(dirb) + '/**', exclude=False, is_path=True),
        ExcludeFilter(name=str(dira) + '/**', exclude=True, is_path=True),
        ExcludeFilter(name='**/foo/*.py', exclude=False, is_path=True),
        ExcludeFilter(name='bar.foo', exclude=False, is_path=False),
        ExcludeFilter(name='bar', exclude=True, is_path=False),
    ]



'''
This module should be run to recreate the files that we generate automatically
(i.e.: modules that shouldn't be traced and cython .pyx)
'''

from __future__ import print_function

import os
import struct
import re


def is_python_64bit():
    return (struct.calcsize('P') == 8)


root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))


def get_cython_contents(filename):
    if filename.endswith('.pyc'):
        filename = filename[:-1]

    state = 'regular'

    replacements = []

    new_contents = []
    with open(filename, 'r') as stream:
        for line in stream:
            strip = line.strip()
            if state == 'regular':
                if strip == '# IFDEF CYTHON':
                    state = 'cython'

                    new_contents.append('%s -- DONT EDIT THIS FILE (it is automatically generated)\n' % line.replace('\n', '').replace('\r', ''))
                    continue

                new_contents.append(line)

            elif state == 'cython':
                if strip == '# ELSE':
                    state = 'nocython'
                    new_contents.append(line)
                    continue

                elif strip == '# ENDIF':
                    state = 'regular'
                    new_contents.append(line)
                    continue

                if strip == '#':
                    continue

                assert strip.startswith('# '), 'Line inside # IFDEF CYTHON must start with "# ". Found: %s' % (strip,)
                strip = strip.replace('# ', '', 1).strip()

                if strip.startswith('cython_inline_constant:'):
                    strip = strip.replace('cython_inline_constant:', '')
                    word_to_replace, replacement = strip.split('=')
                    replacements.append((word_to_replace.strip(), replacement.strip()))
                    continue

                line = line.replace('# ', '', 1)
                new_contents.append(line)

            elif state == 'nocython':
                if strip == '# ENDIF':
                    state = 'regular'
                    new_contents.append(line)
                    continue
                new_contents.append('# %s' % line)

    assert state == 'regular', 'Error: # IFDEF CYTHON found without # ENDIF'

    ret = ''.join(new_contents)

    for (word_to_replace, replacement) in replacements:
        ret = re.sub(r"\b%s\b" % (word_to_replace,), replacement, ret)

    return ret


def _generate_cython_from_files(target, modules):
    contents = ['''from __future__ import print_function

# Important: Autogenerated file.

# DO NOT edit manually!
# DO NOT edit manually!
''']

    for mod in modules:
        contents.append(get_cython_contents(mod.__file__))

    with open(target, 'w') as stream:
        stream.write(''.join(contents))


def generate_dont_trace_files():
    template = '''# Important: Autogenerated file.

# DO NOT edit manually!
# DO NOT edit manually!

LIB_FILE = 1
PYDEV_FILE = 2

DONT_TRACE_DIRS = {
%(pydev_dirs)s
}

DONT_TRACE = {
    # commonly used things from the stdlib that we don't want to trace
    'Queue.py':LIB_FILE,
    'queue.py':LIB_FILE,
    'socket.py':LIB_FILE,
    'weakref.py':LIB_FILE,
    '_weakrefset.py':LIB_FILE,
    'linecache.py':LIB_FILE,
    'threading.py':LIB_FILE,
    'dis.py':LIB_FILE,

    # things from pydev that we don't want to trace
%(pydev_files)s
}

# if we try to trace io.py it seems it can get halted (see http://bugs.python.org/issue4716)
DONT_TRACE['io.py'] = LIB_FILE

# Don't trace common encodings too
DONT_TRACE['cp1252.py'] = LIB_FILE
DONT_TRACE['utf_8.py'] = LIB_FILE
DONT_TRACE['codecs.py'] = LIB_FILE
'''

    pydev_files = []
    pydev_dirs = []

    exclude_dirs = [
        '.git',
        '.settings',
        'build',
        'build_tools',
        'dist',
        'pydevd.egg-info',
        'pydevd_attach_to_process',
        'pydev_sitecustomize',
        'stubs',
        'tests',
        'tests_mainloop',
        'tests_python',
        'tests_runfiles',
        'test_pydevd_reload',
        'third_party',
        '__pycache__',
        'pydev_ipython',
        'vendored',
        '.mypy_cache',
        'pydevd.egg-info',
    ]

    for root, dirs, files in os.walk(root_dir):

        for d in dirs:
            if 'pydev' in d and d != 'pydevd.egg-info':
                # print(os.path.join(root, d))
                pydev_dirs.append("    '%s': PYDEV_FILE," % (d,))

        for d in exclude_dirs:
            try:
                dirs.remove(d)
            except:
                pass

        for f in files:
            if f.endswith('.py'):
                if f not in (
                    '__init__.py',
                    'runfiles.py',
                    'pydev_coverage.py',
                    'pydev_pysrc.py',
                    'setup.py',
                    'setup_pydevd_cython.py',
                    'interpreterInfo.py',
                    'conftest.py',
                    ):
                    pydev_files.append("    '%s': PYDEV_FILE," % (f,))

    contents = template % (dict(
        pydev_files='\n'.join(sorted(pydev_files)),
        pydev_dirs='\n'.join(sorted(pydev_dirs)),
    ))
    assert 'pydevd.py' in contents
    assert 'pydevd_dont_trace.py' in contents
    with open(os.path.join(root_dir, '_pydevd_bundle', 'pydevd_dont_trace_files.py'), 'w') as stream:
        stream.write(contents)


def remove_if_exists(f):
    try:
        if os.path.exists(f):
            os.remove(f)
    except:
        import traceback;traceback.print_exc()


def generate_cython_module():
    remove_if_exists(os.path.join(root_dir, '_pydevd_bundle', 'pydevd_cython.pyx'))

    target = os.path.join(root_dir, '_pydevd_bundle', 'pydevd_cython.pyx')
    curr = os.environ.get('PYDEVD_USE_CYTHON')
    try:
        os.environ['PYDEVD_USE_CYTHON'] = 'NO'

        from _pydevd_bundle import pydevd_additional_thread_info_regular
        from _pydevd_bundle import pydevd_frame, pydevd_trace_dispatch_regular
        _generate_cython_from_files(target, [pydevd_additional_thread_info_regular, pydevd_frame, pydevd_trace_dispatch_regular])
    finally:
        if curr is None:
            del os.environ['PYDEVD_USE_CYTHON']
        else:
            os.environ['PYDEVD_USE_CYTHON'] = curr


if __name__ == '__main__':
    generate_dont_trace_files()
    generate_cython_module()



import threading
event_set = False
inner_started = False

def method():
    global inner_started
    inner_started = True
    while not event_set:
        import time
        time.sleep(.1)
        
t = threading.Thread(target=method)
t.start()
while not inner_started:
    import time
    time.sleep(.1)

print('break here')
event_set = True
t.join()
print('TEST SUCEEDED!')


"""
Helper main which starts a server which listens in a socket.

This makes it easier to start it under a debugger.

The extension must have the "robot.language-server.tcp-port" setting set to 1456.
"""
import sys


def dev_main():
    import os.path

    try:
        import robotframework_ls
    except ImportError:
        # Automatically add it to the path if __main__ is being executed.
        sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        import robotframework_ls  # @UnusedImport
    robotframework_ls.import_robocorp_ls_core()

    from robotframework_ls.__main__ import main

    sys.argv = [
        sys.argv[0],
        "-vv",
        "--tcp",
        "--port=1456",
        # "--log-file=c:/temp/robotlog.log",
    ]

    main()


if __name__ == "__main__":
    while True:
        print("--- Starting new dev session ---")
        dev_main()



from robot.api.parsing import ModelTransformer, EmptyLine, Comment, Token

from robotidy.decorators import check_start_end_line
from robotidy.exceptions import InvalidParameterValueError


class OrderSettings(ModelTransformer):
    """
    Order settings like [Arguments], [Setup], [Return] inside Keywords and Test Cases.

    Keyword settings [Documentation], [Tags], [Timeout], [Arguments] are put before keyword body and
    settings like [Teardown], [Return] are moved to the end of the keyword:

       *** Keywords ***
        Keyword
            [Teardown]  Keyword
            [Return]  ${value}
            [Arguments]  ${arg}
            [Documentation]  this is
            ...    doc
            [Tags]  sanity
            Pass

    To:

       *** Keywords ***
        Keyword
            [Documentation]  this is
            ...    doc
            [Tags]  sanity
            [Arguments]  ${arg}
            Pass
            [Teardown]  Keyword
            [Return]  ${value}

    Test case settings [Documentation], [Tags], [Template], [Timeout], [Setup] are put before test case body and
    [Teardown] is moved to the end of test case.

    Default order can be changed using following parameters:
      - ``keyword_before = documentation,tags,timeout,arguments``
      - ``keyword_after = teardown,return``
      - ``test_before = documentation,tags,template,timeout,setup
      - ``test_after = teardown

    Not all settings names need to be passed to given parameter. Missing setting names are not ordered. Example::

        robotidy --configure OrderSettings:keyword_before=:keyword_after=

    It will order only test cases because all setting names for keywords are missing.

    Supports global formatting params: ``--startline`` and ``--endline``.

    See https://robotidy.readthedocs.io/en/latest/transformers/OrderSettings.html for more examples.
    """

    def __init__(
        self,
        keyword_before: str = None,
        keyword_after: str = None,
        test_before: str = None,
        test_after: str = None,
    ):
        (
            self.keyword_before,
            self.keyword_after,
            self.test_before,
            self.test_after,
        ) = self.parse_order(keyword_before, keyword_after, test_before, test_after)
        self.keyword_settings = {*self.keyword_before, *self.keyword_after}
        self.test_settings = {*self.test_before, *self.test_after}

    def get_order(self, order, default, name_map):
        if order is None:
            return default
        if not order:
            return []
        parts = order.lower().split(",")
        try:
            return [name_map[part] for part in parts]
        except KeyError:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "order",
                order,
                f"Custom order should be provided in comma separated list with valid setting names:\n{sorted(name_map.keys())}",
            )

    def parse_order(self, keyword_before, keyword_after, test_before, test_after):
        keyword_order_before = (
            Token.DOCUMENTATION,
            Token.TAGS,
            Token.TIMEOUT,
            Token.ARGUMENTS,
        )
        keyword_order_after = (
            Token.TEARDOWN,
            Token.RETURN,
        )
        testcase_order_before = (
            Token.DOCUMENTATION,
            Token.TAGS,
            Token.TEMPLATE,
            Token.TIMEOUT,
            Token.SETUP,
        )
        testcase_order_after = (Token.TEARDOWN,)
        keyword_map = {
            "documentation": Token.DOCUMENTATION,
            "tags": Token.TAGS,
            "timeout": Token.TIMEOUT,
            "arguments": Token.ARGUMENTS,
            "return": Token.RETURN,
            "teardown": Token.TEARDOWN,
        }
        test_map = {
            "documentation": Token.DOCUMENTATION,
            "tags": Token.TAGS,
            "timeout": Token.TIMEOUT,
            "template": Token.TEMPLATE,
            "setup": Token.SETUP,
            "teardown": Token.TEARDOWN,
        }
        return (
            self.get_order(keyword_before, keyword_order_before, keyword_map),
            self.get_order(keyword_after, keyword_order_after, keyword_map),
            self.get_order(test_before, testcase_order_before, test_map),
            self.get_order(test_after, testcase_order_after, test_map),
        )

    @check_start_end_line
    def visit_Keyword(self, node):  # noqa
        return self.order_settings(node, self.keyword_settings, self.keyword_before, self.keyword_after)

    @check_start_end_line
    def visit_TestCase(self, node):  # noqa
        return self.order_settings(node, self.test_settings, self.test_before, self.test_after)

    def order_settings(self, node, setting_types, before, after):
        if not node.body:
            return node
        settings = dict()
        not_settings, trailing_after = [], []
        after_seen = False
        # when after_seen is set to True then all statements go to trailing_after and last non data
        # will be appended after tokens defined in `after` set (like [Return])
        comment = []
        for child in node.body:
            if isinstance(child, Comment) and child.lineno == node.lineno:
                comment.append(child)
            elif getattr(child, "type", "invalid") in setting_types:
                after_seen = after_seen or child.type in after
                settings[child.type] = child
            elif after_seen:
                trailing_after.append(child)
            else:
                not_settings.append(child)
        # comments after last data statement are considered as comment outside body
        trailing_non_data = []
        while trailing_after and isinstance(trailing_after[-1], (EmptyLine, Comment)):
            trailing_non_data.insert(0, trailing_after.pop())
        not_settings += trailing_after
        node.body = (
            comment
            + self.add_in_order(before, settings)
            + not_settings
            + self.add_in_order(after, settings)
            + trailing_non_data
        )
        return node

    @staticmethod
    def add_in_order(order, settings_in_node):
        return [settings_in_node[token_type] for token_type in order if token_type in settings_in_node]



"""
Note: we have our own logging framework so that it doesn't conflict with the
regular logging framework usage (because when doing a robotframework launch it
disrupts the existing loggers).

Usage:

configure_logger('LSP', log_level, log_file)
log = get_logger(__name__)

...

log.debug('debug')
log.info('info')
log.critical('critical')
log.exception('error')

Note that we have 3 levels:
0: show only critical/exceptions
1: show info
2: show debug
"""
import os.path
import traceback
import threading
import sys
from datetime import datetime
from robocorp_ls_core.protocols import ILog
from typing import Dict
from robocorp_ls_core.constants import NULL

name_to_logger: Dict[str, ILog] = {}


def _as_str(s):
    if isinstance(s, bytes):
        return s.decode("utf-8", "replace")
    return s


MAX_LOG_MSG_SIZE = 2000
try:
    MAX_LOG_MSG_SIZE = int(os.environ.get("MAX_LOG_MSG_SIZE", MAX_LOG_MSG_SIZE))
except:
    pass


class _LogConfig(object):

    __slots__ = ["_lock", "__stream", "prefix", "log_level", "_log_file", "pid"]

    def __init__(self):
        self._lock = threading.Lock()
        self.__stream = None

        self.prefix = ""
        self.log_level = 0
        self.log_file = None
        self.pid = os.getpid()

    @property
    def log_file(self):
        return self._log_file

    @log_file.setter
    def log_file(self, log_file):
        with self._lock:
            if log_file is None:
                self._log_file = None
                self.__stream = None

            elif isinstance(log_file, str):
                self._log_file = log_file
                self.__stream = None

            else:
                assert hasattr(log_file, "write")
                self._log_file = None
                self.__stream = log_file

    @property
    def _stream(self):
        stream = self.__stream
        if stream is None:

            # open it on demand
            with self._lock:
                log_file = self._log_file
                if self.__stream is None:
                    stream = sys.stderr
                    if log_file:
                        stream = open(log_file, "w")
                    self.__stream = stream
                    return stream

        return stream

    def close_logging_streams(self):
        with self._lock:
            if self.__stream is not None:
                self.__stream.write("-- Closing logging streams --")
                self.__stream.close()
            self.__stream = NULL

    def report(self, logger_name, show_stacktrace, levelname, message, trim):
        if trim:
            msg_len = len(message)
            if msg_len > MAX_LOG_MSG_SIZE:
                message = f"{message[:MAX_LOG_MSG_SIZE-200]} ... <trimmed {msg_len} to {MAX_LOG_MSG_SIZE}> ... {message[-200:]}"

        log_format = (
            self.prefix
            + ": %(asctime)s UTC pid: %(process)d - %(threadname)s - %(levelname)s - %(name)s\n%(message)s\n\n"
        )
        msg = log_format % {
            "asctime": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "process": self.pid,
            "threadname": threading.current_thread().name,
            "levelname": levelname,
            "name": logger_name,
            "message": message,
        }
        try:
            # Get stream without using the lock to avoid deadlocks.
            stream = self._stream
            with self._lock:
                stream.write(msg)
                if show_stacktrace:
                    traceback.print_exc(file=stream)
                stream.flush()
        except:
            pass  # Never fail when logging.


_log_config = _LogConfig()


def close_logging_streams():
    _log_config.close_logging_streams()


class _Logger(object):
    def __init__(self, name):
        self.name = name

    def critical(self, msg="", *args):
        if _log_config.log_level >= 0:
            self._report("CRITICAL", False, False, msg, *args)

    def exception(self, msg="", *args):
        if _log_config.log_level >= 0:
            self._report("EXCEPTION", True, False, msg, *args)

    def info(self, msg="", *args):
        if _log_config.log_level >= 1:
            self._report("INFO", False, True, msg, *args)

    def debug(self, msg="", *args):
        if _log_config.log_level >= 2:
            self._report("DEBUG", False, True, msg, *args)

    warn = warning = info
    error = exception

    @property
    def level(self):
        # Note: return a level compatible with the logging.
        import logging

        log_level = _log_config.log_level
        if log_level >= 2:
            return logging.DEBUG
        if log_level >= 1:
            return logging.INFO
        return logging.ERROR

    def _report(self, levelname, show_stacktrace, trim, msg="", *args):
        msg = _as_str(msg)
        if args:
            args = tuple(_as_str(arg) for arg in args)
            try:
                message = msg % args
            except:
                message = "%s - %s" % (msg, args)
        else:
            message = msg

        _log_config.report(self.name, show_stacktrace, levelname, message, trim)


def get_log_level():
    return _log_config.log_level


def get_log_file():
    return _log_config.log_file


def get_logger(name: str) -> ILog:
    """
    Use as:
        log = get_logger(__name__)
    """
    name = _as_str(name)
    try:
        return name_to_logger[name]
    except:
        name_to_logger[name] = _Logger(name)
    return name_to_logger[name]


def _configure_logger(prefix, log_level, log_file):
    _log_config.prefix = prefix
    _log_config.log_level = log_level
    _log_config.log_file = log_file


def _current_config():
    return _log_config.prefix, _log_config.log_level, _log_config.log_file


class _RestoreCtxManager(object):
    def __init__(self, config_to_restore):
        self._config_to_restore = config_to_restore

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        _configure_logger(*self._config_to_restore)


def configure_logger(postfix, log_level, log_file):
    """
    :param log_file:
        - If None, get target file from env var.
        - If empty string log to stderr.

    :note: If used as a context manager it'll revert to the previous
           configuration on `__exit__`.
    """

    prev_config = _current_config()

    if log_file:
        if isinstance(log_file, str):
            try:
                log_file = os.path.expanduser(log_file)
                log_file = os.path.realpath(os.path.abspath(log_file))
                dirname = os.path.dirname(log_file)
                basename = os.path.basename(log_file)
                try:
                    os.makedirs(dirname)
                except:
                    pass  # Ignore error if it already exists.

                name, ext = os.path.splitext(basename)
                log_file = os.path.join(
                    dirname, name + "." + postfix + "." + str(os.getpid()) + ext
                )
            except:
                log_file = None
                # Don't fail when trying to setup logging, just show the exception.
                traceback.print_exc()

    _configure_logger(postfix, log_level, log_file)
    return _RestoreCtxManager(prev_config)


def log_args_and_python(log, argv, module):
    log.info("Arguments: %s", argv)
    log.info(
        "Python: %s - lsp: %s (%s) - platform: %s - sys.prefix: %s - sys.executable: %s",
        sys.version,
        getattr(module, "__version__", "<unable to get __version__>"),
        module,
        sys.platform,
        sys.prefix,
        sys.executable,
    )



import gevent


def foo():
    print('Running in foo')
    gevent.sleep(0)
    print('Explicit context switch to foo again')


def bar():
    print('Explicit context to bar')
    gevent.sleep(0)  # break here
    print('Implicit context switch back to bar')


if __name__ == '__main__':
    gevent.joinall([
        gevent.spawn(foo),
        gevent.spawn(bar),
    ])
    print('TEST SUCEEDED')



from typing import List, Optional, Dict, Iterator

from robocorp_ls_core.lsp import LocationTypedDict, RangeTypedDict
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IRobotDocument,
    IKeywordFound,
    IVariablesCollector,
    IVariableFound,
    cast_to_keyword_definition,
    AbstractVariablesCollector,
)
import typing
from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.basic import isinstance_name, normalize_filename


log = get_logger(__name__)


def matches_source(s1: str, s2: str) -> bool:
    if s1 == s2:
        return True

    return normalize_filename(s1) == normalize_filename(s2)


class _VariableDefinitionsCollector(AbstractVariablesCollector):
    def __init__(self, robot_string_matcher):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher

        self.robot_string_matcher: RobotStringMatcher = robot_string_matcher
        self.matches: List[IVariableFound] = []

    def accepts(self, variable_name):
        return self.robot_string_matcher.is_same_variable_name(variable_name)

    def on_variable(self, variable_found: IVariableFound):
        self.matches.append(variable_found)

    def __typecheckself__(self) -> None:
        _: IVariablesCollector = check_implements(self)


def iter_variable_references_in_doc(
    completion_context: ICompletionContext,
    doc: IRobotDocument,
    normalized_name: str,
) -> Iterator[RangeTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.ast_utils import create_range_from_token
    from robotframework_ls.impl.variable_completions import collect_variables

    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robocorp_ls_core.lsp import PositionTypedDict

    robot_string_matcher = RobotStringMatcher(normalized_name)

    ast = doc.get_ast()
    if ast is not None:
        for node_info in ast_utils.iter_variable_references(ast):
            completion_context.check_cancelled()
            if robot_string_matcher.is_same_variable_name(node_info.token.value):
                yield create_range_from_token(node_info.token)

        collector = _VariableDefinitionsCollector(robot_string_matcher)
        collect_variables(completion_context, collector, only_current_doc=True)

        variable: IVariableFound
        for variable in collector.matches:
            start: PositionTypedDict = {
                "line": variable.lineno,
                "character": variable.col_offset,
            }
            end: PositionTypedDict = {
                "line": variable.lineno,
                "character": variable.end_col_offset,
            }
            code_lens_range: RangeTypedDict = {"start": start, "end": end}
            yield code_lens_range


def iter_keyword_references_in_doc(
    completion_context: ICompletionContext,
    doc: IRobotDocument,
    normalized_name: str,
    keyword_found: Optional[IKeywordFound],
) -> Iterator[RangeTypedDict]:
    """
    :param keyword_found: if given, we'll match if the definition actually
    maps to the proper place (if not given, we'll just match based on the name
    without verifying if the definition is the same).
    """
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.find_definition import find_definition
    from robotframework_ls.impl.completion_context import CompletionContext
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    ast = doc.get_ast()
    if ast is not None:
        # Dict with normalized name -> whether it was found or not previously.
        found_in_this_doc: Dict[str, bool] = {}

        # Ok, we have the document, now, load the usages.
        for keyword_usage_info in ast_utils.iter_keyword_usage_tokens(
            ast, collect_args_as_keywords=True
        ):
            completion_context.check_cancelled()
            keword_name_possibly_dotted = keyword_usage_info.name
            found_dot_in_usage = "." in keword_name_possibly_dotted
            if found_dot_in_usage:
                keword_name_not_dotted = keword_name_possibly_dotted.split(".")[-1]
            else:
                keword_name_not_dotted = keword_name_possibly_dotted

            if normalize_robot_name(keword_name_not_dotted) == normalized_name:
                found_once_in_this_doc = found_in_this_doc.get(
                    keword_name_possibly_dotted
                )
                token = keyword_usage_info.token

                line = token.lineno - 1

                if keyword_found is not None:
                    if found_once_in_this_doc is None:
                        # Verify if it's actually the same one (not one defined in
                        # a different place with the same name).

                        new_ctx = CompletionContext(
                            doc,
                            line,
                            token.col_offset,
                            workspace=completion_context.workspace,
                            config=completion_context.config,
                            monitor=completion_context.monitor,
                        )
                        definitions = find_definition(new_ctx)
                        for definition in definitions:
                            found = matches_source(
                                definition.source, keyword_found.source
                            )

                            if found:
                                found_once_in_this_doc = found_in_this_doc[
                                    keword_name_possibly_dotted
                                ] = True
                                break
                        else:
                            found_once_in_this_doc = found_in_this_doc[
                                keword_name_possibly_dotted
                            ] = False
                            continue

                    if not found_once_in_this_doc:
                        continue

                token = keyword_usage_info.token
                if found_dot_in_usage:
                    # We need to create a new token because we just want to match the name part.
                    col_offset = token.col_offset + (
                        len(keword_name_possibly_dotted) - len(keword_name_not_dotted)
                    )
                    end_col_offset = token.end_col_offset
                else:
                    col_offset = token.col_offset
                    end_col_offset = token.end_col_offset

                # Ok, we found it, let's add it to the result.
                yield {
                    "start": {
                        "line": line,
                        "character": col_offset,
                    },
                    "end": {
                        "line": line,
                        "character": end_col_offset,
                    },
                }


def references(
    completion_context: ICompletionContext, include_declaration: bool
) -> List[LocationTypedDict]:

    token_info = completion_context.get_current_token()
    if token_info is None:
        return []

    keyword_found: IKeywordFound
    if token_info.token.type == token_info.token.KEYWORD_NAME:
        if isinstance_name(token_info.node, "KeywordName"):
            from robotframework_ls.impl.find_definition import find_keyword_definition

            definitions = find_keyword_definition(completion_context, token_info)
            if definitions:
                for definition in definitions:
                    as_keyword_definition = cast_to_keyword_definition(definition)
                    if as_keyword_definition:
                        keyword_found = as_keyword_definition.keyword_found
                        return _references_for_keyword_found(
                            completion_context, keyword_found, include_declaration
                        )

    current_keyword_definition_and_usage_info = (
        completion_context.get_current_keyword_definition_and_usage_info()
    )
    if current_keyword_definition_and_usage_info is not None:
        completion_context.monitor.check_cancelled()
        keyword_definition, _usage_info = current_keyword_definition_and_usage_info

        keyword_found = keyword_definition.keyword_found
        return _references_for_keyword_found(
            completion_context, keyword_found, include_declaration
        )

    return []


def _references_for_keyword_found(
    completion_context: ICompletionContext,
    keyword_found: IKeywordFound,
    include_declaration: bool,
):
    from robocorp_ls_core import uris
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    ret: List[LocationTypedDict] = []

    normalized_name = normalize_robot_name(keyword_found.keyword_name)
    # Ok, we have the keyword definition, now, we must actually look for the
    # references...
    if include_declaration:
        ret.append(
            {
                "uri": uris.from_fs_path(keyword_found.source),
                "range": {
                    "start": {
                        "line": keyword_found.lineno,
                        "character": keyword_found.col_offset,
                    },
                    "end": {
                        "line": keyword_found.end_lineno,
                        "character": keyword_found.end_col_offset,
                    },
                },
            }
        )

    from robotframework_ls.impl.workspace_symbols import iter_symbols_caches

    for symbols_cache in iter_symbols_caches(
        None, completion_context, force_all_docs_in_workspace=True, timeout=999999
    ):
        completion_context.check_cancelled()
        if symbols_cache.has_keyword_usage(normalized_name):
            doc: Optional[IRobotDocument] = symbols_cache.get_doc()
            if doc is None:
                uri = symbols_cache.get_uri()
                if uri is None:
                    continue

                doc = typing.cast(
                    Optional[IRobotDocument],
                    completion_context.workspace.get_document(
                        doc_uri=uri, accept_from_file=True
                    ),
                )

                if doc is None:
                    log.debug(
                        "Unable to load document for getting references with uri: %s",
                        uri,
                    )
                    continue

            ref_range: RangeTypedDict
            for ref_range in iter_keyword_references_in_doc(
                completion_context, doc, normalized_name, keyword_found
            ):
                ret.append({"uri": doc.uri, "range": ref_range})

    return ret



from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.basic import is_process_alive, log_and_silence_errors
import itertools
from functools import partial
import os
import sys
import threading
from typing import Any, Dict, Optional
from robocorp_ls_core.protocols import ActionResultDict, IConfig
from robocorp_ls_core.options import DEFAULT_TIMEOUT, USE_TIMEOUTS, NO_TIMEOUT

log = get_logger(__name__)
_next_id = partial(next, itertools.count(0))


def create_server_socket(host, port):
    try:
        import socket as socket_module

        server = socket_module.socket(
            socket_module.AF_INET, socket_module.SOCK_STREAM, socket_module.IPPROTO_TCP
        )
        if sys.platform == "win32":
            server.setsockopt(
                socket_module.SOL_SOCKET, socket_module.SO_EXCLUSIVEADDRUSE, 1
            )
        else:
            server.setsockopt(socket_module.SOL_SOCKET, socket_module.SO_REUSEADDR, 1)

        server.bind((host, port))
    except Exception:
        server.close()  # i.e.: close (we just accept 1 connection).
        raise

    return server


class RfInterpreterServerManager:
    def __init__(
        self,
        verbose: int = 0,
        base_log_file: str = "",
        on_interpreter_message=None,
        uri: str = "",
    ):
        from robotframework_interactive.server.rf_interpreter_ls_config import (
            RfInterpreterRobotConfig,
        )
        from robocorp_ls_core import uris

        assert uri
        self._uri = uri
        self._filename = uris.to_fs_path(uri)
        self._lock_api_client = threading.RLock()
        self._server_process = None
        self._log_extension = ".rf_interpreter"
        self._disposed = False
        # The config allows clients to set the python executable/env.
        self._config: IConfig = RfInterpreterRobotConfig()

        self._verbose = verbose
        self._base_log_file = base_log_file
        self._on_interpreter_message = on_interpreter_message

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def config(self) -> IConfig:
        return self._config

    @config.setter
    def config(self, config: IConfig):
        with self._lock_api_client:
            self._config.update(config.get_full_settings())

    def _get_python_executable(self) -> str:
        with self._lock_api_client:
            from robotframework_interactive.server.rf_interpreter_ls_config import (
                OPTION_ROBOT_PYTHON_EXECUTABLE,
            )

            config = self._config
            python_exe = sys.executable
            if config is not None:
                python_exe = config.get_setting(
                    OPTION_ROBOT_PYTHON_EXECUTABLE, str, default=python_exe
                )
            else:
                log.warning(f"self._config not set in {self.__class__}")
            return python_exe

    def _get_environ(self) -> Dict[str, str]:
        from robotframework_interactive.server.rf_interpreter_ls_config import (
            OPTION_ROBOT_PYTHON_ENV,
            OPTION_ROBOT_PYTHONPATH,
        )

        with self._lock_api_client:
            config = self._config
            env = os.environ.copy()

            env.pop("PYTHONPATH", "")
            env.pop("PYTHONHOME", "")
            env.pop("VIRTUAL_ENV", "")

            if config is not None:
                env_in_settings = config.get_setting(
                    OPTION_ROBOT_PYTHON_ENV, dict, default={}
                )
                for key, val in env_in_settings.items():
                    env[str(key)] = str(val)

                pythonpath_entries = config.get_setting(
                    OPTION_ROBOT_PYTHONPATH, list, []
                )
                if pythonpath_entries:
                    # if robot.pythonpath is defined, append those entries to
                    # the PYTHONPATH env variable when starting the interactive
                    # console.
                    current_pythonpath = env.get("PYTHONPATH", "")
                    if not current_pythonpath:
                        env["PYTHONPATH"] = os.pathsep.join(pythonpath_entries)
                    else:
                        existing = set(current_pythonpath.split(os.pathsep))
                        env["PYTHONPATH"] = (
                            current_pythonpath
                            + os.pathsep
                            + os.pathsep.join(
                                (
                                    str(x)
                                    for x in pythonpath_entries
                                    if x not in existing
                                )
                            )
                        )
            else:
                log.warning("self._config not set in %s" % (self.__class__,))
            return env

    def _get_api_client(self) -> Any:
        with self._lock_api_client:
            server_process = self._server_process

            if server_process is not None:
                # If someone killed it, dispose of internal references
                # and create a new process.
                if not is_process_alive(server_process.pid):
                    self._dispose_server_process()

            if self._disposed:
                log.info("Robot Framework Interpreter server already disposed.")
                return None

            if server_process is None:
                try:
                    from robotframework_interactive.server.rf_interpreter__main__ import (
                        start_server_process,
                    )
                    from robocorp_ls_core.jsonrpc.streams import (
                        JsonRpcStreamWriter,
                        JsonRpcStreamReader,
                    )
                    from robotframework_interactive.server.rf_interpreter_client import (
                        RfInterpreterApiClient,
                    )

                    args = []
                    if self._verbose:
                        args.append("-" + "v" * int(self._verbose))
                    if self._base_log_file:
                        log_id = _next_id()
                        # i.e.: use a log id in case we create more than one in the
                        # same session.
                        if log_id == 0:
                            args.append(
                                "--log-file="
                                + self._base_log_file
                                + self._log_extension
                            )
                        else:
                            args.append(
                                "--log-file="
                                + self._base_log_file
                                + (".%s" % (log_id,))
                                + self._log_extension
                            )

                    python_exe = self._get_python_executable()
                    environ = self._get_environ()
                    connect_event = threading.Event()

                    s = create_server_socket(host="127.0.0.1", port=0)
                    import socket as socket_module

                    new_socket: Optional[socket_module.socket] = None
                    connect_event = threading.Event()

                    def wait_for_connection():
                        nonlocal new_socket
                        try:
                            s.settimeout(
                                DEFAULT_TIMEOUT if USE_TIMEOUTS else NO_TIMEOUT
                            )
                            s.listen(1)
                            new_socket, _addr = s.accept()
                            log.info("Connection accepted")
                        except:
                            log.exception("Server did not connect.")
                        finally:
                            connect_event.set()
                            s.close()

                    t = threading.Thread(target=wait_for_connection)
                    t.start()

                    # Now, we're listening, let's start up the interpreter to connect back.
                    _, port = s.getsockname()
                    args.append("--tcp")
                    args.append("--host")
                    args.append("127.0.0.1")
                    args.append("--port")
                    args.append(str(port))

                    cwd = os.path.dirname(self._filename)
                    if not os.path.isdir(cwd):
                        raise AssertionError(f"CWD passed is not a directory: {cwd}")

                    server_process = start_server_process(
                        args=args, python_exe=python_exe, env=environ, cwd=cwd
                    )

                    self._server_process = server_process

                    connect_event.wait()
                    if new_socket is None:
                        raise RuntimeError(
                            "Timed out while waiting for interpreter server to connect."
                        )

                    read_from = new_socket.makefile("rb")
                    write_to = new_socket.makefile("wb")

                    w = JsonRpcStreamWriter(write_to, sort_keys=True)
                    r = JsonRpcStreamReader(read_from)

                    api = self._rf_interpreter_api_client = RfInterpreterApiClient(
                        w,
                        r,
                        server_process,
                        on_interpreter_message=self._on_interpreter_message,
                    )

                    log.debug(
                        "Initializing rf interpreter api... (this pid: %s, api pid: %s).",
                        os.getpid(),
                        server_process.pid,
                    )
                    api.initialize(process_id=os.getpid())

                except Exception as e:
                    if server_process is None:
                        log.exception(
                            "Error starting rf interpreter server api (server_process=None)."
                        )
                    else:
                        exitcode = server_process.poll()
                        if exitcode is not None:
                            # Note: only read() if the process exited.
                            log.exception(
                                "Error starting rf interpreter server api. Exit code: %s Base exception: %s. Stderr: %s",
                                exitcode,
                                e,
                                server_process.stderr.read(),
                            )
                        else:
                            log.exception(
                                "Error (%s) starting rf interpreter server api (still running). Base exception: %s.",
                                exitcode,
                                e,
                            )
                    self._dispose_server_process()
                finally:
                    if server_process is not None:
                        log.debug(
                            "Server api (%s) created pid: %s", self, server_process.pid
                        )
                    else:
                        log.debug("server_process == None in _get_api_client()")

            return self._rf_interpreter_api_client

    @log_and_silence_errors(log)
    def _dispose_server_process(self):
        from robocorp_ls_core.basic import kill_process_and_subprocesses

        with self._lock_api_client:
            try:
                log.debug("Dispose server process.")
                if self._server_process is not None:
                    if is_process_alive(self._server_process.pid):
                        kill_process_and_subprocesses(self._server_process.pid)
            finally:
                self._disposed = True
                self._server_process = None
                self._rf_interpreter_api_client = None

    def interpreter_start(self, uri: str) -> ActionResultDict:
        api = self._get_api_client()
        if api is not None:
            return api.interpreter_start(uri)
        return {
            "success": False,
            "message": "Unable to start Robot Framework Interpreter server api.",
            "result": None,
        }

    @property
    def waiting_input(self):
        api = self._get_api_client()
        if api is not None:
            return api.waiting_input
        return False

    def interpreter_evaluate(self, code: str) -> ActionResultDict:
        api = self._get_api_client()
        if api is not None:
            return api.interpreter_evaluate(code)
        return {
            "success": False,
            "message": "Robot Framework Interpreter server api not available.",
            "result": None,
        }

    def interpreter_compute_evaluate_text(
        self, code: str, target_type: str = "evaluate"
    ) -> ActionResultDict:
        """
        :param target_type:
            'evaluate': means that the target is an evaluation with the given code.
                This implies that the current code must be changed to make sense
                in the given context.

            'completions': means that the target is a code-completion
                This implies that the current code must be changed to include
                all previous successful evaluations so that the code-completion
                contains the full information up to the current point.
        """
        api = self._get_api_client()
        if api is not None:
            return api.interpreter_compute_evaluate_text(code, target_type)
        return {
            "success": False,
            "message": "Robot Framework Interpreter server api not available.",
            "result": None,
        }

    def interpreter_stop(self):
        api = self._get_api_client()
        if api is not None:
            try:
                return api.interpreter_stop()
            finally:
                # After a stop, also dispose the process. It can't be reused
                # (a new manager must be created).
                self._dispose_server_process()
        return {
            "success": False,
            "message": "Unable to stop Robot Framework Interpreter server api.",
            "result": None,
        }



if __name__ == '__main__':
    import subprocess
    import sys
    import os
    import _debugger_case_remote_2
    args = sys.argv[1:]
    port = int(args.pop(0))

    access_token = None
    client_access_token = None
    while args:
        if args[0] == '--access-token':
            access_token = args[1]
            args = args[2:]
        elif args[0] == '--client-access-token':
            client_access_token = args[1]
            args = args[2:]
        else:
            raise AssertionError('Unable to handle args: %s' % (sys.argv[1:]))

    root_dirname = os.path.dirname(os.path.dirname(__file__))

    if root_dirname not in sys.path:
        sys.path.append(root_dirname)

    import pydevd

    print('before pydevd.settrace')
    sys.stdout.flush()
    pydevd.settrace(
        port=port,
        patch_multiprocessing=True,
        access_token=access_token,
        client_access_token=client_access_token,
    )
    print('after pydevd.settrace')
    sys.stdout.flush()
    f = _debugger_case_remote_2.__file__
    if f.endswith('.pyc'):
        f = f[:-1]
    elif f.endswith('$py.class'):
        f = f[:-len('$py.class')] + '.py'
    print('before call')
    sys.stdout.flush()
    subprocess.check_call([sys.executable, '-u', f])
    print('after call')
    sys.stdout.flush()



'''
Note: run test with manage.py test my_app

This is mostly for experimenting.

The actual code used is mostly a copy of this that lives in `django_debug.py`.
'''

from django.test import SimpleTestCase


def collect_lines_for_django_template(template_contents):
    from django import template
    t = template.Template(template_contents)
    return _collect_valid_lines_in_django_template(t)


def _collect_valid_lines_in_django_template(template):
    lines = set()
    for node in _iternodes(template.nodelist):
        lineno = _get_lineno(node)
        if lineno is not None:
            lines.add(lineno)
    return lines


def _get_lineno(node):
    if hasattr(node, 'token') and hasattr(node.token, 'lineno'):
        return node.token.lineno
    return None


def _iternodes(nodelist):
    for node in nodelist:
        yield node

        try:
            children = node.child_nodelists
        except:
            pass
        else:
            for attr in children:
                nodelist = getattr(node, attr, None)
                if nodelist:
                    # i.e.: yield from _iternodes(nodelist)
                    for node in _iternodes(nodelist):
                        yield node


class MyTest(SimpleTestCase):

    def test_something(self):
        template_contents = '''{% if entries %}
    <ul>
    {% for entry in entries %}
        {% for entry in entries2 %}
            <li>
                {{ entry.key }}
                :
                {{ entry.val }}
            </li>
        {% endfor %}
    {% endfor %}
    </ul>
{% else %}
    <p>No entries are available.</p>
{% endif %}'''

        self.assertEqual(
            collect_lines_for_django_template(template_contents),
            {1, 3, 4, 6, 8, 10, 11, 13}
        )



import os
import pytest

import time

from contextlib import contextmanager
from tests_python.debugger_unittest import IS_PY36_OR_GREATER, IS_CPYTHON
from tests_python.debug_constants import TEST_CYTHON

pytest_plugins = [
    str('tests_python.debugger_fixtures'),
]

pytestmark = pytest.mark.skipif(not IS_PY36_OR_GREATER or not IS_CPYTHON or not TEST_CYTHON, reason='Requires CPython >= 3.6')


@pytest.fixture
def case_setup_force_frame_eval(case_setup):

    def get_environ(writer):
        env = os.environ.copy()
        env['PYDEVD_USE_FRAME_EVAL'] = 'YES'
        env['PYDEVD_USE_CYTHON'] = 'YES'
        return env

    original_test_file = case_setup.test_file

    @contextmanager
    def test_file(*args, **kwargs):
        kwargs.setdefault('get_environ', get_environ)
        with original_test_file(*args, **kwargs) as writer:
            yield writer

    case_setup.test_file = test_file
    return case_setup


def test_step_and_resume(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case_simple_calls.py') as writer:
        writer.write_add_breakpoint(10, 'Method2')
        writer.write_add_breakpoint(2, 'Method1')
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()
        assert hit.suspend_type == 'frame_eval'
        assert hit.line == 10

        writer.write_step_over(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit('108')

        assert hit.line == 11
        # we use tracing debugger while stepping
        assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 2
        # we enable frame evaluation debugger after "Resume" command
        assert hit.suspend_type == "frame_eval"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_step_return(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case56.py') as writer:
        writer.write_add_breakpoint(2, 'Call2')
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()

        assert hit.suspend_type == "frame_eval"
        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_step_return(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit('109')

        assert hit.line == 8
        # Step return uses temporary breakpoint, so we use tracing debugger
        assert hit.suspend_type == "trace"

        writer.write_step_in(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit('107')

        # goes to line 4 in jython (function declaration line)
        assert hit.line in (4, 5)
        # we use tracing debugger for stepping
        assert hit.suspend_type == "trace", 'Expected suspend type to be "trace", but was: %s' % hit.suspend_type

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_break_while_running(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case3.py') as writer:
        writer.write_make_initial_run()
        time.sleep(.5)
        breakpoint_id = writer.write_add_breakpoint(4, '')

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 4
        # we use tracing debugger if breakpoint was added while running
        assert hit.suspend_type == "trace"

        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()
        assert hit.line == 4
        # we still use tracing debugger
        assert hit.suspend_type == "trace"

        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_remove_breakpoint(breakpoint_id)

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_exc_break(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case_simple_calls.py') as writer:
        writer.write_add_breakpoint(10, 'Method2')
        writer.write_add_exception_breakpoint_with_policy('IndexError', "1", "0", "0")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 10
        # we use tracing debugger if there are exception breakpoints
        assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_exc_break_while_running(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case_simple_calls.py') as writer:
        writer.write_add_breakpoint(10, 'Method2')
        writer.write_add_breakpoint(2, 'Method1')
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit('111')

        assert hit.line == 10
        # we use tracing debugger if there are exception breakpoints
        assert hit.suspend_type == "frame_eval"

        writer.write_add_exception_breakpoint_with_policy('IndexError', "1", "0", "0")

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 2
        # we use tracing debugger if exception break was added
        assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_termination_exc_break(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case_simple_calls.py') as writer:
        writer.write_add_breakpoint(10, 'Method2')
        writer.write_add_exception_breakpoint_with_policy('IndexError', "0", "1", "0")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=10)

        # we can use frame evaluation with exception breakpoint with "On termination" suspend policy
        assert hit.suspend_type == "frame_eval"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_frame_eval_whitebox_test(case_setup_force_frame_eval):
    from tests_python.debugger_unittest import CMD_STEP_INTO, CMD_STEP_RETURN, CMD_STEP_OVER

    with case_setup_force_frame_eval.test_file('_debugger_case_frame_eval.py') as writer:
        line_on_global = writer.get_line_index_with_content('break on global')
        writer.write_add_breakpoint(line_on_global, '')
        writer.write_add_breakpoint(writer.get_line_index_with_content('break on check_with_trace'), 'None')
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=line_on_global)
        assert hit.suspend_type == "frame_eval"

        writer.write_step_over(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(reason=CMD_STEP_OVER)

        writer.write_step_in(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(
            line=writer.get_line_index_with_content('check_step_in_then_step_return') + 1, reason=CMD_STEP_INTO)
        assert hit.name == 'check_step_in_then_step_return'

        writer.write_step_return(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(reason=CMD_STEP_RETURN)
        assert hit.name == '<module>'

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_frame_eval_change_breakpoints(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file('_debugger_case_change_breaks.py') as writer:
        break1_line = writer.get_line_index_with_content('break 1')
        break2_line = writer.get_line_index_with_content('break 2')

        break2_id = writer.write_add_breakpoint(break2_line, 'None')
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        assert hit.suspend_type == "frame_eval"

        writer.write_remove_breakpoint(break2_id)
        writer.write_add_breakpoint(break1_line, 'None')
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break1_line)
        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_generator_code_cache(case_setup_force_frame_eval):

    with case_setup_force_frame_eval.test_file('_debugger_case_yield_from.py') as writer:
        break1_line = writer.get_line_index_with_content('break1')
        writer.write_add_breakpoint(break1_line)
        break2_line = writer.get_line_index_with_content('break2')
        writer.write_add_breakpoint(break2_line)
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=break1_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True



from collections import namedtuple
from contextlib import contextmanager
import json
from urllib.parse import quote, quote_plus, unquote_plus

import re
import socket
import subprocess
import threading
import time
import traceback
from tests_python.debug_constants import *

from _pydev_bundle import pydev_localhost, pydev_log

# Note: copied (don't import because we want it to be independent on the actual code because of backward compatibility).
CMD_RUN = 101
CMD_LIST_THREADS = 102
CMD_THREAD_CREATE = 103
CMD_THREAD_KILL = 104
CMD_THREAD_SUSPEND = 105
CMD_THREAD_RUN = 106
CMD_STEP_INTO = 107
CMD_STEP_OVER = 108
CMD_STEP_RETURN = 109
CMD_GET_VARIABLE = 110
CMD_SET_BREAK = 111
CMD_REMOVE_BREAK = 112
CMD_EVALUATE_EXPRESSION = 113
CMD_GET_FRAME = 114
CMD_EXEC_EXPRESSION = 115
CMD_WRITE_TO_CONSOLE = 116
CMD_CHANGE_VARIABLE = 117
CMD_RUN_TO_LINE = 118
CMD_RELOAD_CODE = 119
CMD_GET_COMPLETIONS = 120

# Note: renumbered (conflicted on merge)
CMD_CONSOLE_EXEC = 121
CMD_ADD_EXCEPTION_BREAK = 122
CMD_REMOVE_EXCEPTION_BREAK = 123
CMD_LOAD_SOURCE = 124
CMD_ADD_DJANGO_EXCEPTION_BREAK = 125
CMD_REMOVE_DJANGO_EXCEPTION_BREAK = 126
CMD_SET_NEXT_STATEMENT = 127
CMD_SMART_STEP_INTO = 128
CMD_EXIT = 129
CMD_SIGNATURE_CALL_TRACE = 130

CMD_SET_PY_EXCEPTION = 131
CMD_GET_FILE_CONTENTS = 132
CMD_SET_PROPERTY_TRACE = 133
# Pydev debug console commands
CMD_EVALUATE_CONSOLE_EXPRESSION = 134
CMD_RUN_CUSTOM_OPERATION = 135
CMD_GET_BREAKPOINT_EXCEPTION = 136
CMD_STEP_CAUGHT_EXCEPTION = 137
CMD_SEND_CURR_EXCEPTION_TRACE = 138
CMD_SEND_CURR_EXCEPTION_TRACE_PROCEEDED = 139
CMD_IGNORE_THROWN_EXCEPTION_AT = 140
CMD_ENABLE_DONT_TRACE = 141
CMD_SHOW_CONSOLE = 142

CMD_GET_ARRAY = 143
CMD_STEP_INTO_MY_CODE = 144
CMD_GET_CONCURRENCY_EVENT = 145
CMD_SHOW_RETURN_VALUES = 146

CMD_GET_THREAD_STACK = 152
CMD_THREAD_DUMP_TO_STDERR = 153  # This is mostly for unit-tests to diagnose errors on ci.
CMD_STOP_ON_START = 154
CMD_GET_EXCEPTION_DETAILS = 155
CMD_PYDEVD_JSON_CONFIG = 156

CMD_THREAD_SUSPEND_SINGLE_NOTIFICATION = 157
CMD_THREAD_RESUME_SINGLE_NOTIFICATION = 158

CMD_STEP_OVER_MY_CODE = 159
CMD_STEP_RETURN_MY_CODE = 160

CMD_SET_PY_EXCEPTION = 161
CMD_SET_PATH_MAPPING_JSON = 162

CMD_GET_SMART_STEP_INTO_VARIANTS = 163  # XXX: PyCharm has 160 for this (we're currently incompatible anyways).

CMD_REDIRECT_OUTPUT = 200
CMD_GET_NEXT_STATEMENT_TARGETS = 201
CMD_SET_PROJECT_ROOTS = 202

CMD_AUTHENTICATE = 205

CMD_VERSION = 501
CMD_RETURN = 502
CMD_SET_PROTOCOL = 503
CMD_ERROR = 901

REASON_CAUGHT_EXCEPTION = CMD_STEP_CAUGHT_EXCEPTION
REASON_UNCAUGHT_EXCEPTION = CMD_ADD_EXCEPTION_BREAK
REASON_STOP_ON_BREAKPOINT = CMD_SET_BREAK
REASON_THREAD_SUSPEND = CMD_THREAD_SUSPEND
REASON_STEP_INTO = CMD_STEP_INTO
REASON_STEP_INTO_MY_CODE = CMD_STEP_INTO_MY_CODE
REASON_STOP_ON_START = CMD_STOP_ON_START
REASON_STEP_RETURN = CMD_STEP_RETURN
REASON_STEP_RETURN_MY_CODE = CMD_STEP_RETURN_MY_CODE
REASON_STEP_OVER = CMD_STEP_OVER
REASON_STEP_OVER_MY_CODE = CMD_STEP_OVER_MY_CODE

# Always True (because otherwise when we do have an error, it's hard to diagnose).
SHOW_WRITES_AND_READS = True
SHOW_OTHER_DEBUG_INFO = True
SHOW_STDOUT = True

import platform

IS_CPYTHON = platform.python_implementation() == 'CPython'
IS_IRONPYTHON = platform.python_implementation() == 'IronPython'
IS_JYTHON = platform.python_implementation() == 'Jython'
IS_PYPY = platform.python_implementation() == 'PyPy'
IS_APPVEYOR = os.environ.get('APPVEYOR', '') in ('True', 'true', '1')

try:
    from thread import start_new_thread
except ImportError:
    from _thread import start_new_thread  # @UnresolvedImport

Hit = namedtuple('Hit', 'thread_id, frame_id, line, suspend_type, name, file')


def overrides(method):
    '''
    Helper to check that one method overrides another (redeclared in unit-tests to avoid importing pydevd).
    '''

    def wrapper(func):
        if func.__name__ != method.__name__:
            msg = "Wrong @override: %r expected, but overwriting %r."
            msg = msg % (func.__name__, method.__name__)
            raise AssertionError(msg)

        if func.__doc__ is None:
            func.__doc__ = method.__doc__

        return func

    return wrapper


TIMEOUT = 20

try:
    TimeoutError = TimeoutError  # @ReservedAssignment
except NameError:

    class TimeoutError(RuntimeError):  # @ReservedAssignment
        pass


def wait_for_condition(condition, msg=None, timeout=TIMEOUT, sleep=.05):
    curtime = time.time()
    while True:
        if condition():
            break
        if time.time() - curtime > timeout:
            error_msg = 'Condition not reached in %s seconds' % (timeout,)
            if msg is not None:
                error_msg += '\n'
                if callable(msg):
                    error_msg += msg()
                else:
                    error_msg += str(msg)

            raise TimeoutError(error_msg)
        time.sleep(sleep)


class IgnoreFailureError(RuntimeError):
    pass


#=======================================================================================================================
# ReaderThread
#=======================================================================================================================
class ReaderThread(threading.Thread):

    MESSAGES_TIMEOUT = 15

    def __init__(self, sock):
        threading.Thread.__init__(self)
        self.name = 'Test Reader Thread'
        try:
            from queue import Queue
        except ImportError:
            from Queue import Queue

        self.daemon = True
        self._buffer = b''
        self.sock = sock
        self._queue = Queue()
        self._kill = False
        self.accept_xml_messages = True

    def set_messages_timeout(self, timeout):
        self.MESSAGES_TIMEOUT = timeout

    def get_next_message(self, context_message, timeout=None):
        if timeout is None:
            timeout = self.MESSAGES_TIMEOUT
        try:
            msg = self._queue.get(block=True, timeout=timeout)
        except:
            raise TimeoutError('No message was written in %s seconds. Error message:\n%s' % (timeout, context_message,))
        else:
            frame = sys._getframe().f_back.f_back
            frame_info = ''
            while frame:
                if not frame.f_code.co_name.startswith('test_'):
                    frame = frame.f_back
                    continue

                if frame.f_code.co_filename.endswith('debugger_unittest.py'):
                    frame = frame.f_back
                    continue

                stack_msg = ' --  File "%s", line %s, in %s\n' % (frame.f_code.co_filename, frame.f_lineno, frame.f_code.co_name)
                if 'run' == frame.f_code.co_name:
                    frame_info = stack_msg  # Ok, found the writer thread 'run' method (show only that).
                    break
                frame_info += stack_msg
                frame = frame.f_back
                # Just print the first which is not debugger_unittest.py
                break

            frame = None
            sys.stdout.write('Message returned in get_next_message(): %s --  ctx: %s, asked at:\n%s\n' % (unquote_plus(unquote_plus(msg)), context_message, frame_info))

        if not self.accept_xml_messages:
            if '<xml' in msg:
                raise AssertionError('Xml messages disabled. Received: %s' % (msg,))
        return msg

    def _read(self, size):
        while True:
            buffer_len = len(self._buffer)
            if buffer_len == size:
                ret = self._buffer
                self._buffer = b''
                return ret

            if buffer_len > size:
                ret = self._buffer[:size]
                self._buffer = self._buffer[size:]
                return ret

            r = self.sock.recv(max(size - buffer_len, 1024))
            if not r:
                return b''
            self._buffer += r

    def _read_line(self):
        while True:
            i = self._buffer.find(b'\n')
            if i != -1:
                i += 1  # Add the newline to the return
                ret = self._buffer[:i]
                self._buffer = self._buffer[i:]
                return ret
            else:
                r = self.sock.recv(1024)
                if not r:
                    return b''
                self._buffer += r

    def run(self):
        try:
            content_len = -1

            while not self._kill:
                line = self._read_line()

                if not line:
                    break

                if SHOW_WRITES_AND_READS:
                    show_line = line
                    show_line = line.decode('utf-8')

                    print('%s Received %s' % (self.name, show_line,))

                if line.startswith(b'Content-Length:'):
                    content_len = int(line.strip().split(b':', 1)[1])
                    continue

                if content_len != -1:
                    # If we previously received a content length, read until a '\r\n'.
                    if line == b'\r\n':
                        json_contents = self._read(content_len)
                        content_len = -1

                        if len(json_contents) == 0:
                            self.handle_except()
                            return  # Finished communication.

                        msg = json_contents
                        msg = msg.decode('utf-8')
                        print('Test Reader Thread Received %s' % (msg,))
                        self._queue.put(msg)

                    continue
                else:
                    # No content len, regular line-based protocol message (remove trailing new-line).
                    if line.endswith(b'\n\n'):
                        line = line[:-2]

                    elif line.endswith(b'\n'):
                        line = line[:-1]

                    elif line.endswith(b'\r'):
                        line = line[:-1]

                    msg = line
                    msg = msg.decode('utf-8')
                    print('Test Reader Thread Received %s' % (msg,))
                    self._queue.put(msg)

        except:
            pass  # ok, finished it
        finally:
            # When the socket from pydevd is closed the client should shutdown to notify
            # it acknowledged it.
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
            except:
                pass
            try:
                self.sock.close()
            except:
                pass

    def do_kill(self):
        self._kill = True
        if hasattr(self, 'sock'):
            from socket import SHUT_RDWR
            try:
                self.sock.shutdown(SHUT_RDWR)
            except:
                pass
            try:
                self.sock.close()
            except:
                pass
            delattr(self, 'sock')


def read_process(stream, buffer, debug_stream, stream_name, finish):
    while True:
        line = stream.readline()
        if not line:
            break

        line = line.decode('utf-8', errors='replace')

        if SHOW_STDOUT:
            debug_stream.write('%s: %s' % (stream_name, line,))
        buffer.append(line)

        if finish[0]:
            return


def start_in_daemon_thread(target, args):
    t0 = threading.Thread(target=target, args=args)
    t0.daemon = True
    t0.start()


class DebuggerRunner(object):

    def __init__(self, tmpdir):
        if tmpdir is not None:
            self.pydevd_debug_file = os.path.join(str(tmpdir), 'pydevd_debug_file_%s.txt' % (os.getpid(),))
        else:
            self.pydevd_debug_file = None

    def get_command_line(self):
        '''
        Returns the base command line (i.e.: ['python.exe', '-u'])
        '''
        raise NotImplementedError

    def add_command_line_args(self, args):
        writer = self.writer
        port = int(writer.port)

        localhost = pydev_localhost.get_localhost()
        ret = [
            writer.get_pydevd_file(),
            '--DEBUG_RECORD_SOCKET_READS',
        ]

        if not IS_PY36_OR_GREATER or not IS_CPYTHON or not TEST_CYTHON:
            # i.e.: in frame-eval mode we support native threads, whereas
            # on other cases we need the qt monkeypatch.
            ret += ['--qt-support']

        ret += [
            '--client',
            localhost,
            '--port',
            str(port),
        ]

        if writer.IS_MODULE:
            ret += ['--module']

        ret += ['--file'] + writer.get_command_line_args()
        ret = writer.update_command_line_args(ret)  # Provide a hook for the writer
        return args + ret

    @contextmanager
    def check_case(self, writer_class, wait_for_port=True, wait_for_initialization=True):
        try:
            if callable(writer_class):
                writer = writer_class()
            else:
                writer = writer_class
            try:
                writer.start()
                if wait_for_port:
                    wait_for_condition(lambda: hasattr(writer, 'port'))
                self.writer = writer

                args = self.get_command_line()

                args = self.add_command_line_args(args)

                if SHOW_OTHER_DEBUG_INFO:
                    print('executing: %s' % (' '.join(args),))

                with self.run_process(args, writer) as dct_with_stdout_stder:
                    try:
                        if not wait_for_initialization:
                            # The use-case for this is that the debugger can't even start-up in this
                            # scenario, as such, sleep a bit so that the output can be collected.
                            time.sleep(1)
                        elif wait_for_port:
                            wait_for_condition(lambda: writer.finished_initialization)
                    except TimeoutError:
                        sys.stderr.write('Timed out waiting for initialization\n')
                        sys.stderr.write('stdout:\n%s\n\nstderr:\n%s\n' % (
                            ''.join(dct_with_stdout_stder['stdout']),
                            ''.join(dct_with_stdout_stder['stderr']),
                        ))
                        raise
                    finally:
                        writer.get_stdout = lambda: ''.join(dct_with_stdout_stder['stdout'])
                        writer.get_stderr = lambda: ''.join(dct_with_stdout_stder['stderr'])

                    yield writer
            finally:
                writer.do_kill()
                writer.log = []

            stdout = dct_with_stdout_stder['stdout']
            stderr = dct_with_stdout_stder['stderr']
            writer.additional_output_checks(''.join(stdout), ''.join(stderr))
        except IgnoreFailureError:
            sys.stderr.write('Test finished with ignored failure.\n')
            return

    def create_process(self, args, writer):
        env = writer.get_environ() if writer is not None else None
        if env is None:
            env = os.environ.copy()

        if self.pydevd_debug_file:
            env['PYDEVD_DEBUG'] = 'True'
            env['PYDEVD_DEBUG_FILE'] = self.pydevd_debug_file
            print('Logging to: %s' % (self.pydevd_debug_file,))
        process = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            cwd=writer.get_cwd() if writer is not None else '.',
            env=env,
        )
        return process

    @contextmanager
    def run_process(self, args, writer):
        process = self.create_process(args, writer)
        writer.process = process
        stdout = []
        stderr = []
        finish = [False]
        dct_with_stdout_stder = {}
        fail_with_message = False

        try:
            start_in_daemon_thread(read_process, (process.stdout, stdout, sys.stdout, 'stdout', finish))
            start_in_daemon_thread(read_process, (process.stderr, stderr, sys.stderr, 'stderr', finish))

            if SHOW_OTHER_DEBUG_INFO:
                print('Both processes started')

            # polls can fail (because the process may finish and the thread still not -- so, we give it some more chances to
            # finish successfully).
            initial_time = time.time()
            shown_intermediate = False
            dumped_threads = False

            dct_with_stdout_stder['stdout'] = stdout
            dct_with_stdout_stder['stderr'] = stderr
            try:
                yield dct_with_stdout_stder
            except:
                fail_with_message = True
                # Let's print the actuayl exception here (it doesn't appear properly on Python 2 and
                # on Python 3 it's hard to find because pytest output is too verbose).
                sys.stderr.write('***********\n')
                sys.stderr.write('***********\n')
                sys.stderr.write('***********\n')
                traceback.print_exc()
                sys.stderr.write('***********\n')
                sys.stderr.write('***********\n')
                sys.stderr.write('***********\n')
                raise

            if not writer.finished_ok:
                self.fail_with_message(
                    "The thread that was doing the tests didn't finish successfully (writer.finished_ok = True not set).",
                    stdout,
                    stderr,
                    writer
                )

            while True:
                if process.poll() is not None:
                    if writer.EXPECTED_RETURNCODE != 'any':
                        expected_returncode = writer.EXPECTED_RETURNCODE
                        if not isinstance(expected_returncode, (list, tuple)):
                            expected_returncode = (expected_returncode,)

                        if process.returncode not in expected_returncode:
                            self.fail_with_message('Expected process.returncode to be %s. Found: %s' % (
                                writer.EXPECTED_RETURNCODE, process.returncode), stdout, stderr, writer)
                    break
                else:
                    if writer is not None:
                        if writer.FORCE_KILL_PROCESS_WHEN_FINISHED_OK:
                            process.kill()
                            continue

                        if not shown_intermediate and (time.time() - initial_time > (TIMEOUT / 3.)):  # 1/3 of timeout
                            print('Warning: writer thread exited and process still did not (%.2f seconds elapsed).' % (time.time() - initial_time,))
                            shown_intermediate = True

                        if time.time() - initial_time > ((TIMEOUT / 3.) * 2.):  # 2/3 of timeout
                            if not dumped_threads:
                                dumped_threads = True
                                # It still didn't finish. Ask for a thread dump
                                # (we'll be able to see it later on the test output stderr).
                                try:
                                    writer.write_dump_threads()
                                except:
                                    traceback.print_exc()

                        if time.time() - initial_time > TIMEOUT:  # timed out
                            process.kill()
                            time.sleep(.2)
                            self.fail_with_message(
                                "The other process should've exited but still didn't (%.2f seconds timeout for process to exit)." % (time.time() - initial_time,),
                                stdout, stderr, writer
                            )
                time.sleep(.2)

            if writer is not None:
                if not writer.FORCE_KILL_PROCESS_WHEN_FINISHED_OK:
                    if stdout is None:
                        self.fail_with_message(
                            "The other process may still be running -- and didn't give any output.", stdout, stderr, writer)

                    check = 0
                    while not writer.check_test_suceeded_msg(stdout, stderr):
                        check += 1
                        if check == 50:
                            self.fail_with_message("TEST SUCEEDED not found.", stdout, stderr, writer)
                        time.sleep(.1)

        except TimeoutError:
            msg = 'TimeoutError'
            try:
                writer.write_dump_threads()
            except:
                msg += ' (note: error trying to dump threads on timeout).'
            time.sleep(.2)
            self.fail_with_message(msg, stdout, stderr, writer)
        except Exception as e:
            if fail_with_message:
                self.fail_with_message(str(e), stdout, stderr, writer)
            else:
                raise
        finally:
            try:
                if process.poll() is None:
                    process.kill()
            except:
                traceback.print_exc()
            finish[0] = True

    def fail_with_message(self, msg, stdout, stderr, writerThread):
        log_contents = ''
        for f in pydev_log.list_log_files(self.pydevd_debug_file):
            if os.path.exists(f):
                with open(f, 'r') as stream:
                    log_contents += '\n-------------------- %s ------------------\n\n' % (f,)
                    log_contents += stream.read()
        msg += ("\n\n===========================\nStdout: \n" + ''.join(stdout) +
            "\n\n===========================\nStderr:" + ''.join(stderr) +
            "\n\n===========================\nWriter Log:\n" + '\n'.join(getattr(writerThread, 'log', [])) +
            "\n\n===========================\nLog:" + log_contents)

        if IS_JYTHON:
            # It seems we have some spurious errors which make Jython tests flaky (on a test run it's
            # not unusual for one test among all the tests to fail with this error on Jython).
            # The usual traceback in this case is:
            #
            # Traceback (most recent call last):
            #   File "/home/travis/build/fabioz/PyDev.Debugger/_pydevd_bundle/pydevd_comm.py", line 287, in _on_run
            #     line = self._read_line()
            #   File "/home/travis/build/fabioz/PyDev.Debugger/_pydevd_bundle/pydevd_comm.py", line 270, in _read_line
            #     r = self.sock.recv(1024)
            #   File "/home/travis/build/fabioz/PyDev.Debugger/_pydevd_bundle/pydevd_comm.py", line 270, in _read_line
            #     r = self.sock.recv(1024)
            #   File "/home/travis/jython/Lib/_socket.py", line 1270, in recv
            #     data, _ = self._get_message(bufsize, "recv")
            #   File "/home/travis/jython/Lib/_socket.py", line 384, in handle_exception
            #     raise _map_exception(jlx)
            # error: [Errno -1] Unmapped exception: java.lang.NullPointerException
            #
            # So, ignore errors in this situation.

            if 'error: [Errno -1] Unmapped exception: java.lang.NullPointerException' in msg:
                raise IgnoreFailureError()
        raise AssertionError(msg)


#=======================================================================================================================
# AbstractWriterThread
#=======================================================================================================================
class AbstractWriterThread(threading.Thread):

    FORCE_KILL_PROCESS_WHEN_FINISHED_OK = False
    IS_MODULE = False
    TEST_FILE = None
    EXPECTED_RETURNCODE = 0

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.process = None  # Set after the process is created.
        self.daemon = True
        self.finished_ok = False
        self.finished_initialization = False
        self._next_breakpoint_id = 0
        self.log = []

    def run(self):
        self.start_socket()

    def check_test_suceeded_msg(self, stdout, stderr):
        return 'TEST SUCEEDED' in ''.join(stdout)

    def update_command_line_args(self, args):
        return args

    def _ignore_stderr_line(self, line):
        if line.startswith((
            'debugger: ',
            '>>',
            '<<',
            'warning: Debugger speedups',
            'pydev debugger: New process is launching',
            'pydev debugger: To debug that process',
            '*** Multiprocess',
            )):
            return True

        for expected in (
            'PyDev console: using IPython',
            'Attempting to work in a virtualenv. If you encounter problems, please',
            'Unable to create basic Accelerated OpenGL',  # Issue loading qt5
            'Core Image is now using the software OpenGL',  # Issue loading qt5
            'XDG_RUNTIME_DIR not set',  # Issue loading qt5
            ):
            if expected in line:
                return True

        if re.match(r'^(\d+)\t(\d)+', line):
            return True

        if IS_JYTHON:
            for expected in (
                'org.python.netty.util.concurrent.DefaultPromise',
                'org.python.netty.util.concurrent.SingleThreadEventExecutor',
                'Failed to submit a listener notification task. Event loop shut down?',
                'java.util.concurrent.RejectedExecutionException',
                'An event executor terminated with non-empty task',
                'java.lang.UnsupportedOperationException',
                "RuntimeWarning: Parent module '_pydevd_bundle' not found while handling absolute import",
                'from _pydevd_bundle.pydevd_additional_thread_info_regular import _current_frames',
                'from _pydevd_bundle.pydevd_additional_thread_info import _current_frames',
                'import org.python.core as PyCore #@UnresolvedImport',
                'from _pydevd_bundle.pydevd_additional_thread_info import set_additional_thread_info',
                "RuntimeWarning: Parent module '_pydevd_bundle._debug_adapter' not found while handling absolute import",
                'import json',

                # Issues with Jython and Java 9.
                'WARNING: Illegal reflective access by org.python.core.PySystemState',
                'WARNING: Please consider reporting this to the maintainers of org.python.core.PySystemState',
                'WARNING: An illegal reflective access operation has occurred',
                'WARNING: Illegal reflective access by jnr.posix.JavaLibCHelper',
                'WARNING: Please consider reporting this to the maintainers of jnr.posix.JavaLibCHelper',
                'WARNING: Use --illegal-access=warn to enable warnings of further illegal reflective access operations',
                'WARNING: All illegal access operations will be denied in a future release',
                ):
                if expected in line:
                    return True

            if line.strip().startswith('at '):
                return True

        return False

    def additional_output_checks(self, stdout, stderr):
        lines_with_error = []
        for line in stderr.splitlines():
            line = line.strip()
            if not line:
                continue
            if not self._ignore_stderr_line(line):
                lines_with_error.append(line)

        if lines_with_error:
            raise AssertionError('Did not expect to have line(s) in stderr:\n\n%s\n\nFull stderr:\n\n%s' % (
                '\n'.join(lines_with_error), stderr))

    def get_environ(self):
        return None

    def get_pydevd_file(self):
        dirname = os.path.dirname(__file__)
        dirname = os.path.dirname(dirname)
        return os.path.abspath(os.path.join(dirname, 'pydevd.py'))

    def get_pydevconsole_file(self):
        dirname = os.path.dirname(__file__)
        dirname = os.path.dirname(dirname)
        return os.path.abspath(os.path.join(dirname, 'pydevconsole.py'))

    def get_line_index_with_content(self, line_content, filename=None):
        '''
        :return the line index which has the given content (1-based).
        '''
        if filename is None:
            filename = self.TEST_FILE
        with open(filename, 'r') as stream:
            for i_line, line in enumerate(stream):
                if line_content in line:
                    return i_line + 1
        raise AssertionError('Did not find: %s in %s' % (line_content, self.TEST_FILE))

    def get_cwd(self):
        return os.path.dirname(self.get_pydevd_file())

    def get_command_line_args(self):
        return [self.TEST_FILE]

    def do_kill(self):
        if hasattr(self, 'server_socket'):
            self.server_socket.close()
            delattr(self, 'server_socket')

        if hasattr(self, 'reader_thread'):
            # if it's not created, it's not there...
            self.reader_thread.do_kill()
            delattr(self, 'reader_thread')

        if hasattr(self, 'sock'):
            self.sock.close()
            delattr(self, 'sock')

        if hasattr(self, 'port'):
            delattr(self, 'port')

    def write_with_content_len(self, msg):
        self.log.append('write: %s' % (msg,))

        if SHOW_WRITES_AND_READS:
            print('Test Writer Thread Written %s' % (msg,))

        if not hasattr(self, 'sock'):
            print('%s.sock not available when sending: %s' % (self, msg))
            return

        if not isinstance(msg, bytes):
            msg = msg.encode('utf-8')

        self.sock.sendall((u'Content-Length: %s\r\n\r\n' % len(msg)).encode('ascii'))
        self.sock.sendall(msg)

    _WRITE_LOG_PREFIX = 'write: '

    def write(self, s):
        from _pydevd_bundle.pydevd_comm import ID_TO_MEANING
        meaning = ID_TO_MEANING.get(re.search(r'\d+', s).group(), '')
        if meaning:
            meaning += ': '

        self.log.append(self._WRITE_LOG_PREFIX + '%s%s' % (meaning, s,))

        if SHOW_WRITES_AND_READS:
            print('Test Writer Thread Written %s%s' % (meaning, s,))
        msg = s + '\n'

        if not hasattr(self, 'sock'):
            print('%s.sock not available when sending: %s' % (self, msg))
            return

        msg = msg.encode('utf-8')

        self.sock.send(msg)

    def get_next_message(self, context_message, timeout=None):
        return self.reader_thread.get_next_message(context_message, timeout=timeout)

    def start_socket(self, port=None):
        assert not hasattr(self, 'port'), 'Socket already initialized.'
        from _pydev_bundle.pydev_localhost import get_socket_name
        if SHOW_WRITES_AND_READS:
            print('start_socket')

        self._sequence = -1
        if port is None:
            socket_name = get_socket_name(close=True)
        else:
            socket_name = (pydev_localhost.get_localhost(), port)
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(socket_name)
        self.port = socket_name[1]
        server_socket.listen(1)
        if SHOW_WRITES_AND_READS:
            print('Waiting in socket.accept()')
        self.server_socket = server_socket
        new_socket, addr = server_socket.accept()
        if SHOW_WRITES_AND_READS:
            print('Test Writer Thread Socket:', new_socket, addr)

        self._set_socket(new_socket)

    def _set_socket(self, new_socket):
        curr_socket = getattr(self, 'sock', None)
        if curr_socket:
            try:
                curr_socket.shutdown(socket.SHUT_WR)
            except:
                pass
            try:
                curr_socket.close()
            except:
                pass

        reader_thread = self.reader_thread = ReaderThread(new_socket)
        self.sock = new_socket
        reader_thread.start()

        # initial command is always the version
        self.write_version()
        self.log.append('start_socket')
        self.finished_initialization = True

    def start_socket_client(self, host, port):
        self._sequence = -1
        if SHOW_WRITES_AND_READS:
            print("Connecting to %s:%s" % (host, port))

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        #  Set TCP keepalive on an open socket.
        #  It activates after 1 second (TCP_KEEPIDLE,) of idleness,
        #  then sends a keepalive ping once every 3 seconds (TCP_KEEPINTVL),
        #  and closes the connection after 5 failed ping (TCP_KEEPCNT), or 15 seconds
        try:
            from socket import IPPROTO_TCP, SO_KEEPALIVE, TCP_KEEPIDLE, TCP_KEEPINTVL, TCP_KEEPCNT
            s.setsockopt(socket.SOL_SOCKET, SO_KEEPALIVE, 1)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 3)
            s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 5)
        except ImportError:
            pass  # May not be available everywhere.

        # 10 seconds default timeout
        timeout = int(os.environ.get('PYDEVD_CONNECT_TIMEOUT', 10))
        s.settimeout(timeout)
        for _i in range(20):
            try:
                s.connect((host, port))
                break
            except:
                time.sleep(.5)  # We may have to wait a bit more and retry (especially on PyPy).
        s.settimeout(None)  # no timeout after connected
        if SHOW_WRITES_AND_READS:
            print("Connected.")
        self._set_socket(s)
        return s

    def next_breakpoint_id(self):
        self._next_breakpoint_id += 1
        return self._next_breakpoint_id

    def next_seq(self):
        self._sequence += 2
        return self._sequence

    def wait_for_new_thread(self):
        # wait for hit breakpoint
        last = ''
        while not '<xml><thread name="' in last or '<xml><thread name="pydevd.' in last:
            last = self.get_next_message('wait_for_new_thread')

        # we have something like <xml><thread name="MainThread" id="12103472" /></xml>
        splitted = last.split('"')
        thread_id = splitted[3]
        return thread_id

    def wait_for_output(self):
        # Something as:
        # <xml><io s="TEST SUCEEDED%2521" ctx="1"/></xml>
        while True:
            msg = self.get_next_message('wait_output')
            if "<xml><io s=" in msg:
                if 'ctx="1"' in msg:
                    ctx = 'stdout'
                elif 'ctx="2"' in msg:
                    ctx = 'stderr'
                else:
                    raise AssertionError('IO message without ctx.')

                msg = unquote_plus(unquote_plus(msg.split('"')[1]))
                return msg, ctx

    def get_current_stack_hit(self, thread_id, **kwargs):
        self.write_get_thread_stack(thread_id)
        msg = self.wait_for_message(CMD_GET_THREAD_STACK)
        return self._get_stack_as_hit(msg, **kwargs)

    def wait_for_single_notification_as_hit(self, reason=REASON_STOP_ON_BREAKPOINT, **kwargs):
        dct = self.wait_for_json_message(CMD_THREAD_SUSPEND_SINGLE_NOTIFICATION)
        assert dct['stop_reason'] == reason

        line = kwargs.pop('line', None)
        file = kwargs.pop('file', None)
        assert not kwargs, 'Unexpected kwargs: %s' % (kwargs,)

        return self.get_current_stack_hit(dct['thread_id'], line=line, file=file)

    def wait_for_breakpoint_hit(self, reason=REASON_STOP_ON_BREAKPOINT, timeout=None, **kwargs):
        '''
        108 is over
        109 is return
        111 is breakpoint

        :param reason: may be the actual reason (int or string) or a list of reasons.
        '''
        # note: those must be passed in kwargs.
        line = kwargs.get('line')
        file = kwargs.get('file')
        name = kwargs.get('name')

        self.log.append('Start: wait_for_breakpoint_hit')
        # wait for hit breakpoint
        if not isinstance(reason, (list, tuple)):
            reason = (reason,)

        def accept_message(last):
            for r in reason:
                if ('stop_reason="%s"' % (r,)) in last:
                    return True

            return False

        msg = self.wait_for_message(accept_message, timeout=timeout)
        return self._get_stack_as_hit(msg, file, line, name)

    def _get_stack_as_hit(self, msg, file=None, line=None, name=None):
        # we have something like <xml><thread id="12152656" stop_reason="111"><frame id="12453120" name="encode" ...
        if len(msg.thread.frame) == 0:
            frame = msg.thread.frame
        else:
            frame = msg.thread.frame[0]
        thread_id = msg.thread['id']
        frame_id = frame['id']
        suspend_type = msg.thread['suspend_type']
        hit_name = frame['name']
        frame_line = int(frame['line'])
        frame_file = frame['file']

        if file is not None:
            assert frame_file.endswith(file), 'Expected hit to be in file %s, was: %s' % (file, frame_file)

        if line is not None:
            assert line == frame_line, 'Expected hit to be in line %s, was: %s (in file: %s)' % (line, frame_line, frame_file)

        if name is not None:
            if not isinstance(name, (list, tuple, set)):
                assert name == hit_name
            else:
                assert hit_name in name

        self.log.append('End(1): wait_for_breakpoint_hit: %s' % (msg.original_xml,))

        return Hit(
            thread_id=thread_id, frame_id=frame_id, line=frame_line, suspend_type=suspend_type, name=hit_name, file=frame_file)

    def wait_for_get_next_statement_targets(self):
        last = ''
        while not '<xml><line>' in last:
            last = self.get_next_message('wait_for_get_next_statement_targets')

        matches = re.finditer(r"(<line>([0-9]*)<\/line>)", last, re.IGNORECASE)
        lines = []
        for _, match in enumerate(matches):
            try:
                lines.append(int(match.group(2)))
            except ValueError:
                pass
        return set(lines)

    def wait_for_custom_operation(self, expected):
        # wait for custom operation response, the response is double encoded
        expected_encoded = quote(quote_plus(expected))
        last = ''
        while not expected_encoded in last:
            last = self.get_next_message('wait_for_custom_operation. Expected (encoded): %s' % (expected_encoded,))

        return True

    def _is_var_in_last(self, expected, last):
        if expected in last:
            return True

        last = unquote_plus(last)
        if expected in last:
            return True

        # We actually quote 2 times on the backend...
        last = unquote_plus(last)
        if expected in last:
            return True

        return False

    def wait_for_multiple_vars(self, expected_vars):
        if not isinstance(expected_vars, (list, tuple)):
            expected_vars = [expected_vars]

        all_found = []
        ignored = []

        while True:
            try:
                last = self.get_next_message('wait_for_multiple_vars: %s' % (expected_vars,))
            except:
                missing = []
                for v in expected_vars:
                    if v not in all_found:
                        missing.append(v)
                raise ValueError('Not Found:\n%s\nNot found messages: %s\nFound messages: %s\nExpected messages: %s\nIgnored messages:\n%s' % (
                    '\n'.join(str(x) for x in missing), len(missing), len(all_found), len(expected_vars), '\n'.join(str(x) for x in ignored)))

            was_message_used = False
            new_expected = []
            for expected in expected_vars:
                found_expected = False
                if isinstance(expected, (tuple, list)):
                    for e in expected:
                        if self._is_var_in_last(e, last):
                            was_message_used = True
                            found_expected = True
                            all_found.append(expected)
                            break
                else:
                    if self._is_var_in_last(expected, last):
                        was_message_used = True
                        found_expected = True
                        all_found.append(expected)

                if not found_expected:
                    new_expected.append(expected)

            expected_vars = new_expected

            if not expected_vars:
                return True

            if not was_message_used:
                ignored.append(last)

    wait_for_var = wait_for_multiple_vars
    wait_for_vars = wait_for_multiple_vars
    wait_for_evaluation = wait_for_multiple_vars

    def write_make_initial_run(self):
        self.write("101\t%s\t" % self.next_seq())
        self.log.append('write_make_initial_run')

    def write_set_protocol(self, protocol):
        self.write("%s\t%s\t%s" % (CMD_SET_PROTOCOL, self.next_seq(), protocol))

    def write_authenticate(self, access_token, client_access_token):
        msg = "%s\t%s\t%s" % (CMD_AUTHENTICATE, self.next_seq(), access_token)
        self.write(msg)

        self.wait_for_message(lambda msg:client_access_token in msg, expect_xml=False)

    def write_version(self):
        from _pydevd_bundle.pydevd_constants import IS_WINDOWS
        self.write("%s\t%s\t1.0\t%s\tID" % (CMD_VERSION, self.next_seq(), 'WINDOWS' if IS_WINDOWS else 'UNIX'))

    def get_main_filename(self):
        return self.TEST_FILE

    def write_show_return_vars(self, show=1):
        self.write("%s\t%s\tCMD_SHOW_RETURN_VALUES\t%s" % (CMD_SHOW_RETURN_VALUES, self.next_seq(), show))

    def write_add_breakpoint(self, line, func='None', filename=None, hit_condition=None, is_logpoint=False, suspend_policy=None, condition=None):
        '''
        :param line: starts at 1
        :param func: if None, may hit in any context, empty string only top level, otherwise must be method name.
        '''
        if filename is None:
            filename = self.get_main_filename()
        breakpoint_id = self.next_breakpoint_id()
        if hit_condition is None and not is_logpoint and suspend_policy is None and condition is None:
            # Format kept for backward compatibility tests
            self.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\tNone\tNone" % (
                CMD_SET_BREAK, self.next_seq(), breakpoint_id, 'python-line', filename, line, func))
        else:
            # Format: breakpoint_id, type, file, line, func_name, condition, expression, hit_condition, is_logpoint, suspend_policy
            self.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tNone\t%s\t%s\t%s" % (
                CMD_SET_BREAK, self.next_seq(), breakpoint_id, 'python-line', filename, line, func, condition, hit_condition, is_logpoint, suspend_policy))
        self.log.append('write_add_breakpoint: %s line: %s func: %s' % (breakpoint_id, line, func))
        return breakpoint_id

    def write_multi_threads_single_notification(self, multi_threads_single_notification):
        self.write_json_config(dict(
            multi_threads_single_notification=multi_threads_single_notification,
        ))

    def write_suspend_on_breakpoint_exception(self, skip_suspend_on_breakpoint_exception, skip_print_breakpoint_exception):
        self.write_json_config(dict(
            skip_suspend_on_breakpoint_exception=skip_suspend_on_breakpoint_exception,
            skip_print_breakpoint_exception=skip_print_breakpoint_exception
        ))

    def write_json_config(self, config_dict):
        self.write("%s\t%s\t%s" % (CMD_PYDEVD_JSON_CONFIG, self.next_seq(),
            json.dumps(config_dict)
        ))

    def write_stop_on_start(self, stop=True):
        self.write("%s\t%s\t%s" % (CMD_STOP_ON_START, self.next_seq(), stop))

    def write_dump_threads(self):
        self.write("%s\t%s\t" % (CMD_THREAD_DUMP_TO_STDERR, self.next_seq()))

    def write_add_exception_breakpoint(self, exception):
        self.write("%s\t%s\t%s" % (CMD_ADD_EXCEPTION_BREAK, self.next_seq(), exception))
        self.log.append('write_add_exception_breakpoint: %s' % (exception,))

    def write_get_current_exception(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_GET_EXCEPTION_DETAILS, self.next_seq(), thread_id))

    def write_set_py_exception_globals(
            self,
            break_on_uncaught,
            break_on_caught,
            skip_on_exceptions_thrown_in_same_context,
            ignore_exceptions_thrown_in_lines_with_ignore_exception,
            ignore_libraries,
            exceptions=()
        ):
        # Only set the globals, others
        self.write("131\t%s\t%s" % (self.next_seq(), '%s;%s;%s;%s;%s;%s' % (
            'true' if break_on_uncaught else 'false',
            'true' if break_on_caught else 'false',
            'true' if skip_on_exceptions_thrown_in_same_context else 'false',
            'true' if ignore_exceptions_thrown_in_lines_with_ignore_exception else 'false',
            'true' if ignore_libraries else 'false',
            ';'.join(exceptions)
        )))
        self.log.append('write_set_py_exception_globals')

    def write_start_redirect(self):
        self.write("%s\t%s\t%s" % (CMD_REDIRECT_OUTPUT, self.next_seq(), 'STDERR STDOUT'))

    def write_set_project_roots(self, project_roots):
        self.write("%s\t%s\t%s" % (CMD_SET_PROJECT_ROOTS, self.next_seq(), '\t'.join(str(x) for x in project_roots)))

    def write_add_exception_breakpoint_with_policy(
            self, exception, notify_on_handled_exceptions, notify_on_unhandled_exceptions, ignore_libraries):
        self.write("%s\t%s\t%s" % (CMD_ADD_EXCEPTION_BREAK, self.next_seq(), '\t'.join(str(x) for x in [
            exception, notify_on_handled_exceptions, notify_on_unhandled_exceptions, ignore_libraries])))
        self.log.append('write_add_exception_breakpoint: %s' % (exception,))

    def write_remove_exception_breakpoint(self, exception):
        self.write('%s\t%s\t%s' % (CMD_REMOVE_EXCEPTION_BREAK, self.next_seq(), exception))

    def write_remove_breakpoint(self, breakpoint_id):
        self.write("%s\t%s\t%s\t%s\t%s" % (
            CMD_REMOVE_BREAK, self.next_seq(), 'python-line', self.get_main_filename(), breakpoint_id))

    def write_change_variable(self, thread_id, frame_id, varname, value):
        self.write("%s\t%s\t%s\t%s\t%s\t%s\t%s" % (
            CMD_CHANGE_VARIABLE, self.next_seq(), thread_id, frame_id, 'FRAME', varname, value))

    def write_get_frame(self, thread_id, frame_id):
        self.write("%s\t%s\t%s\t%s\tFRAME" % (CMD_GET_FRAME, self.next_seq(), thread_id, frame_id))
        self.log.append('write_get_frame')

    def write_get_variable(self, thread_id, frame_id, var_attrs):
        self.write("%s\t%s\t%s\t%s\tFRAME\t%s" % (CMD_GET_VARIABLE, self.next_seq(), thread_id, frame_id, var_attrs))

    def write_step_over(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_OVER, self.next_seq(), thread_id,))

    def write_step_in(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_INTO, self.next_seq(), thread_id,))

    def write_step_in_my_code(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_INTO_MY_CODE, self.next_seq(), thread_id,))

    def write_step_return(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_RETURN, self.next_seq(), thread_id,))

    def write_step_return_my_code(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_RETURN_MY_CODE, self.next_seq(), thread_id,))

    def write_step_over_my_code(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_STEP_OVER_MY_CODE, self.next_seq(), thread_id,))

    def write_suspend_thread(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_THREAD_SUSPEND, self.next_seq(), thread_id,))

    def write_reload(self, module_name):
        self.log.append('write_reload')
        self.write("%s\t%s\t%s" % (CMD_RELOAD_CODE, self.next_seq(), module_name,))

    def write_run_thread(self, thread_id):
        self.log.append('write_run_thread')
        self.write("%s\t%s\t%s" % (CMD_THREAD_RUN, self.next_seq(), thread_id,))

    def write_get_thread_stack(self, thread_id):
        self.log.append('write_get_thread_stack')
        self.write("%s\t%s\t%s" % (CMD_GET_THREAD_STACK, self.next_seq(), thread_id,))

    def write_load_source(self, filename):
        self.log.append('write_load_source')
        self.write("%s\t%s\t%s" % (CMD_LOAD_SOURCE, self.next_seq(), filename,))

    def write_load_source_from_frame_id(self, frame_id):
        from _pydevd_bundle.pydevd_comm_constants import CMD_LOAD_SOURCE_FROM_FRAME_ID
        self.log.append('write_load_source_from_frame_id')
        self.write("%s\t%s\t%s" % (CMD_LOAD_SOURCE_FROM_FRAME_ID, self.next_seq(), frame_id,))

    def write_kill_thread(self, thread_id):
        self.write("%s\t%s\t%s" % (CMD_THREAD_KILL, self.next_seq(), thread_id,))

    def write_set_next_statement(self, thread_id, line, func_name):
        self.write("%s\t%s\t%s\t%s\t%s" % (CMD_SET_NEXT_STATEMENT, self.next_seq(), thread_id, line, func_name,))

    def write_smart_step_into(self, thread_id, line, func_name):
        self.write("%s\t%s\t%s\t%s\t%s" % (CMD_SMART_STEP_INTO, self.next_seq(), thread_id, line, func_name,))

    def write_debug_console_expression(self, locator):
        self.write("%s\t%s\t%s" % (CMD_EVALUATE_CONSOLE_EXPRESSION, self.next_seq(), locator))

    def write_custom_operation(self, locator, style, codeOrFile, operation_fn_name):
        self.write("%s\t%s\t%s||%s\t%s\t%s" % (
            CMD_RUN_CUSTOM_OPERATION, self.next_seq(), locator, style, codeOrFile, operation_fn_name))

    def write_evaluate_expression(self, locator, expression):
        self.write("%s\t%s\t%s\t%s\t1" % (CMD_EVALUATE_EXPRESSION, self.next_seq(), locator, expression))

    def write_enable_dont_trace(self, enable):
        if enable:
            enable = 'true'
        else:
            enable = 'false'
        self.write("%s\t%s\t%s" % (CMD_ENABLE_DONT_TRACE, self.next_seq(), enable))

    def write_get_next_statement_targets(self, thread_id, frame_id):
        self.write("201\t%s\t%s\t%s" % (self.next_seq(), thread_id, frame_id))
        self.log.append('write_get_next_statement_targets')

    def write_list_threads(self):
        seq = self.next_seq()
        self.write("%s\t%s\t" % (CMD_LIST_THREADS, seq))
        return seq

    def wait_for_list_threads(self, seq):
        return self.wait_for_message('502')

    def wait_for_get_thread_stack_message(self):
        return self.wait_for_message(CMD_GET_THREAD_STACK)

    def wait_for_curr_exc_stack(self):
        return self.wait_for_message(CMD_SEND_CURR_EXCEPTION_TRACE)

    def wait_for_json_message(self, accept_message, unquote_msg=True, timeout=None):
        last = self.wait_for_message(accept_message, unquote_msg, expect_xml=False, timeout=timeout)
        json_msg = last.split('\t', 2)[-1]  # We have something as: CMD\tSEQ\tJSON
        if isinstance(json_msg, bytes):
            json_msg = json_msg.decode('utf-8')
        try:
            return json.loads(json_msg)
        except:
            traceback.print_exc()
            raise AssertionError('Unable to parse:\n%s\njson:\n%s' % (last, json_msg))

    def wait_for_message(self, accept_message, unquote_msg=True, expect_xml=True, timeout=None):
        if isinstance(accept_message, (str, int)):
            msg_starts_with = '%s\t' % (accept_message,)

            def accept_message(msg):
                return msg.startswith(msg_starts_with)

        import untangle
        from io import StringIO
        prev = None
        while True:
            last = self.get_next_message('wait_for_message', timeout=timeout)
            if unquote_msg:
                last = unquote_plus(unquote_plus(last))
            if accept_message(last):
                if expect_xml:
                    # Extract xml and return untangled.
                    xml = ''
                    try:
                        xml = last[last.index('<xml>'):]
                        if isinstance(xml, bytes):
                            xml = xml.decode('utf-8')
                        xml = untangle.parse(StringIO(xml))
                    except:
                        traceback.print_exc()
                        raise AssertionError('Unable to parse:\n%s\nxml:\n%s' % (last, xml))
                    ret = xml.xml
                    ret.original_xml = last
                    return ret
                else:
                    return last
            if prev != last:
                print('Ignored message: %r' % (last,))

            prev = last

    def get_frame_names(self, thread_id):
        self.write_get_thread_stack(thread_id)
        msg = self.wait_for_message(CMD_GET_THREAD_STACK)
        if msg.thread.frame:
            frame_names = [frame['name'] for frame in msg.thread.frame]
            return frame_names
        return [msg.thread.frame['name']]

    def get_step_into_variants(self, thread_id, frame_id, start_line, end_line):
        self.write("%s\t%s\t%s\t%s\t%s\t%s" % (CMD_GET_SMART_STEP_INTO_VARIANTS, self.next_seq(), thread_id, frame_id, start_line, end_line))
        msg = self.wait_for_message(CMD_GET_SMART_STEP_INTO_VARIANTS)
        if msg.variant:
            variant_info = [
                (variant['name'], variant['isVisited'], variant['line'], variant['callOrder'], variant['offset'], variant['childOffset'])
                for variant in msg.variant
            ]
            return variant_info
        return []

    def wait_for_thread_join(self, main_thread_id):

        def condition():
            return self.get_frame_names(main_thread_id) in (
                ['wait', 'join', '<module>'],
                ['_wait_for_tstate_lock', 'join', '<module>']
            )

        def msg():
            return 'Found stack: %s' % (self.get_frame_names(main_thread_id),)

        wait_for_condition(condition, msg, timeout=5, sleep=.5)

    def create_request_thread(self, full_url):

        class T(threading.Thread):

            def wait_for_contents(self):
                for _ in range(10):
                    if hasattr(self, 'contents'):
                        break
                    time.sleep(.3)
                else:
                    raise AssertionError('Unable to get contents from server. Url: %s' % (full_url,))
                return self.contents

            def run(self):
                try:
                    from urllib.request import urlopen
                except ImportError:
                    from urllib import urlopen
                for _ in range(10):
                    try:
                        stream = urlopen(full_url)
                        contents = stream.read()
                        contents = contents.decode('utf-8')
                        self.contents = contents
                        break
                    except IOError:
                        continue

        t = T()
        t.daemon = True
        return t


def _get_debugger_test_file(filename):
    ret = os.path.abspath(os.path.join(os.path.dirname(__file__), filename))
    if not os.path.exists(ret):
        ret = os.path.join(os.path.dirname(__file__), 'resources', filename)
    if not os.path.exists(ret):
        raise AssertionError('Expected: %s to exist.' % (ret,))
    return ret


def get_free_port():
    from _pydev_bundle.pydev_localhost import get_socket_name
    return get_socket_name(close=True)[1]
